import numpy as np
import matplotlib.pyplot as plt

from .area_geometry import AreaGeometry


class HalfPlane(AreaGeometry):
    """
    This class represents a half plane in 2d.

    Attributes
    ----------
    centre: numpy.ndarray
        The centre point, shape: (2,)
    n: numpy.ndarray
        The direction vector to the inside, shape: (2,)
    n: numpy.ndarray
        The direction vector orthogonal to n, shape: (2,)

    :group: utils.geom2d

    """

    def __init__(self, centre, n):
        """
        Constructor.

        Parameters
        ----------
        centre: numpy.ndarray
            The centre point, shape: (2,)
        n: numpy.ndarray
            The direction vector to the inside, shape: (2,)

        """
        self.centre = np.array(centre, dtype=np.float64)
        self.n = np.array(n, dtype=np.float64)
        self.n /= np.linalg.norm(self.n)
        self.m = np.array([-self.n[1], self.n[0]], dtype=np.float64)

    def p_min(self):
        """
        Returns minimal (x,y) point.

        Returns
        -------
        p_min: numpy.ndarray
            The minimal (x,y) point, shape = (2,)

        """
        if np.linalg.norm(self.n - np.array([1, 0])) < 1e-13:
            return np.array([self.centre[0], -np.inf], dtype=np.float64)
        if np.linalg.norm(self.n - np.array([-1, 0])) < 1e-13:
            return np.array([-np.inf, -np.inf], dtype=np.float64)
        if np.linalg.norm(self.n - np.array([0, 1])) < 1e-13:
            return np.array([-np.inf, self.centre[1]], dtype=np.float64)
        if np.linalg.norm(self.n - np.array([0, -1])) < 1e-13:
            return np.array([-np.inf, -np.inf], dtype=np.float64)

        return np.array([-np.inf, -np.inf], dtype=np.float64)

    def p_max(self):
        """
        Returns maximal (x,y) point.

        Returns
        -------
        p_min: numpy.ndarray
            The maximal (x,y) point, shape = (2,)

        """
        if np.linalg.norm(self.n - np.array([1, 0])) < 1e-13:
            return np.array([np.inf, np.inf], dtype=np.float64)
        if np.linalg.norm(self.n - np.array([-1, 0])) < 1e-13:
            return np.array([self.centre[0], np.inf], dtype=np.float64)
        if np.linalg.norm(self.n - np.array([0, 1])) < 1e-13:
            return np.array([np.inf, np.inf], dtype=np.float64)
        if np.linalg.norm(self.n - np.array([0, -1])) < 1e-13:
            return np.array([np.inf, self.centre[1]], dtype=np.float64)

        return np.array([np.inf, np.inf], dtype=np.float64)

    def points_distance(self, points, return_nearest=False):
        """
        Calculates point distances wrt boundary.

        Parameters
        ----------
        points: numpy.ndarray
            The probe points, shape (n_points, 2)
        return_nearest: bool
            Flag for return of the nearest point on bundary

        Returns
        -------
        dist: numpy.ndarray
            The smallest distances to the boundary,
            shape: (n_points,)
        p_nearest: numpy.ndarray, optional
            The nearest points on the boundary, if
            return_nearest is True, shape: (n_points, 2)

        """

        deltas = points - self.centre[None, :]
        x = np.einsum("pd,d->p", deltas, self.n)

        if return_nearest:
            y = np.einsum("pd,d->p", deltas, self.m)
            nerst = self.centre[None, :] + y[:, None] * self.m[None, :]
            return np.abs(x), nerst
        else:
            return np.abs(x)

    def points_inside(self, points):
        """
        Tests if points are inside the geometry.

        Parameters
        ----------
        points: numpy.ndarray
            The probe points, shape (n_points, 2)

        Returns
        -------
        inside: numpy.ndarray
            True if point is inside, shape: (n_points,)

        """
        deltas = points - self.centre[None, :]
        x = np.einsum("pd,d->p", deltas, self.n)
        return x >= 0.0

    def add_to_figure(
        self, ax, show_boundary=True, fill_mode=None, pars_boundary={}, pars_distance={}
    ):
        """
        Add image to (x,y) figure.

        Parameters
        ----------
        ax: matplotlib.pyplot.Axis
            The axis object
        show_boundary: bool
            Add the boundary line to the image
        fill_mode: str, optional
            Fill the area. Options:
            dist, dist_inside, dist_outside, inside_<color>,
            outside_<color>
        pars_boundary: dict
            Parameters for boundary plotting command
        pars_distance: dict
            Parameters for distance plotting command

        """
        if show_boundary:
            pars = dict(color="darkblue", linewidth=1)
            pars.update(pars_boundary)

            ax.axline(tuple(self.centre), tuple(self.centre + self.m), **pars)

        super().add_to_figure(
            ax, show_boundary, fill_mode, pars_boundary, pars_distance
        )

    def inverse(self):
        """
        Get the inverted geometry

        Returns
        -------
        inverted: foxes.utils.geom2d.InvertedAreaGeometry
            The inverted geometry

        """
        return HalfPlane(self.centre, -self.n)


if __name__ == "__main__":
    from .circle import Circle

    p0 = [4, 5]
    n = [1.0, 0.3]

    centre = np.array([3.0, 4.0])
    radius = 2.5
    N = 500

    fig, ax = plt.subplots()
    g = Circle(centre, radius) + HalfPlane(p0, n)
    g.add_to_figure(ax, fill_mode="dist_inside")
    plt.show()
    plt.close(fig)

    fig, ax = plt.subplots()
    g.add_to_figure(ax, fill_mode="dist_outside")
    plt.show()
    plt.close(fig)

    fig, ax = plt.subplots()
    g = Circle(centre, radius) - HalfPlane(p0, n)
    g.add_to_figure(ax, fill_mode="dist_inside")
    plt.show()
    plt.close(fig)

    fig, ax = plt.subplots()
    g.add_to_figure(ax, fill_mode="dist_outside")
    plt.show()
    plt.close(fig)

    fig, ax = plt.subplots()
    g = g.inverse()
    g.add_to_figure(ax, fill_mode="dist_inside")
    plt.show()
    plt.close(fig)

    fig, ax = plt.subplots()
    g.add_to_figure(ax, fill_mode="dist_outside")
    plt.show()
    plt.close(fig)



from abc import abstractmethod

from .model import Model


class AxialInductionModel(Model):
    """
    Abstract base class for axial induction models

    :group: core

    """

    @abstractmethod
    def ct2a(self, ct):
        """
        Computes induction from ct

        Parameters
        ----------
        ct: numpy.ndarray or float
            The ct values

        Returns
        -------
        ct: numpy.ndarray or float
            The induction values

        """
        pass



import plotly.io as pio
import io
from PIL import Image


def show_plotly_fig(fig):
    """
    Displays a plotly figure in a window

    Reference:
    https://stackoverflow.com/questions/53570384/plotly-how-to-make-a-standalone-plot-in-a-window

    :group: utils

    """
    buf = io.BytesIO()
    pio.write_image(fig, buf)
    img = Image.open(buf)
    img.show()



from abc import abstractmethod

from .model import Model


class WakeSuperposition(Model):
    """
    Abstract base class for wake superposition models.

    Note that it is a matter of the wake model
    if superposition models are used, or if the
    wake model computes the total wake result by
    other means.

    :group: core

    """

    @abstractmethod
    def add_wake(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        st_sel,
        variable,
        wake_delta,
        wake_model_result,
    ):
        """
        Add a wake delta to previous wake deltas,
        at rotor points.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwnd order
        st_sel: numpy.ndarray of bool
            The selection of targets, shape: (n_states, n_targets)
        variable: str
            The variable name for which the wake deltas applies
        wake_delta: numpy.ndarray
            The original wake deltas, shape:
            (n_states, n_targets, n_tpoints, ...)
        wake_model_result: numpy.ndarray
            The new wake deltas of the selected rotors,
            shape: (n_st_sel, n_tpoints, ...)

        Returns
        -------
        wdelta: numpy.ndarray
            The updated wake deltas, shape:
            (n_states, n_targets, n_tpoints, ...)

        """
        pass

    @abstractmethod
    def calc_final_wake_delta(
        self,
        algo,
        mdata,
        fdata,
        variable,
        amb_results,
        wake_delta,
    ):
        """
        Calculate the final wake delta after adding all
        contributions.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        variable: str
            The variable name for which the wake deltas applies
        amb_results: numpy.ndarray
            The ambient results at targets,
            shape: (n_states, n_targets, n_tpoints)
        wake_delta: numpy.ndarray
            The wake deltas at targets, shape:
            (n_states, n_targets, n_tpoints)

        Returns
        -------
        final_wake_delta: numpy.ndarray
            The final wake delta, which will be added to the ambient
            results by simple plus operation. Shape:
            (n_states, n_targets, n_tpoints)

        """
        pass



import numpy as np
from iwopy import Objective
from scipy.spatial.distance import cdist

import foxes.constants as FC


class OMaxN(Objective):
    """
    Maximal number of turbines objective
    for purely geometrical layouts problems.

    :group: opt.problems.layout.geom_layouts.objectives

    """

    def __init__(self, problem, name="maxN"):
        """
        Constructor.

        Parameters
        ----------
        problem: foxes.opt.FarmOptProblem
            The underlying geometrical layout
            optimization problem
        name: str
            The constraint name

        """
        super().__init__(
            problem,
            name,
            vnames_int=problem.var_names_int(),
            vnames_float=problem.var_names_float(),
        )

    def n_components(self):
        """
        Returns the number of components of the
        function.

        Returns
        -------
        int:
            The number of components.

        """
        return 1

    def maximize(self):
        """
        Returns flag for maximization of each component.

        Returns
        -------
        flags: np.array
            Bool array for component maximization,
            shape: (n_components,)

        """
        return [True]

    def calc_individual(self, vars_int, vars_float, problem_results, cmpnts=None):
        """
        Calculate values for a single individual of the
        underlying problem.

        Parameters
        ----------
        vars_int : np.array
            The integer variable values, shape: (n_vars_int,)
        vars_float : np.array
            The float variable values, shape: (n_vars_float,)
        problem_results : Any
            The results of the variable application
            to the problem
        components : list of int, optional
            The selected components or None for all

        Returns
        -------
        values : np.array
            The component values, shape: (n_sel_components,)

        """
        __, valid = problem_results
        return np.sum(valid)

    def calc_population(self, vars_int, vars_float, problem_results, cmpnts=None):
        """
        Calculate values for all individuals of a population.

        Parameters
        ----------
        vars_int : np.array
            The integer variable values, shape: (n_pop, n_vars_int)
        vars_float : np.array
            The float variable values, shape: (n_pop, n_vars_float)
        problem_results : Any
            The results of the variable application
            to the problem
        components : list of int, optional
            The selected components or None for all

        Returns
        -------
        values : np.array
            The component values, shape: (n_pop, n_sel_components)

        """
        __, valid = problem_results
        return np.sum(valid, axis=1)[:, None]


class OMinN(OMaxN):
    """
    Minimal number of turbines objective
    for purely geometrical layouts problems.

    :group: opt.problems.layout.geom_layouts.objectives

    """

    def __init__(self, problem, name="ominN"):
        """
        Constructor.

        Parameters
        ----------
        problem: foxes.opt.FarmOptProblem
            The underlying geometrical layout
            optimization problem
        name: str
            The constraint name

        """
        super().__init__(problem, name)

    def maximize(self):
        return [False]


class OFixN(Objective):
    """
    Fixed number of turbines objective
    for purely geometrical layouts problems.

    :group: opt.problems.layout.geom_layouts.objectives

    """

    def __init__(self, problem, N, name="ofixN"):
        """
        Constructor.

        Parameters
        ----------
        problem: foxes.opt.FarmOptProblem
            The underlying geometrical layout
            optimization problem
        N: int
            The number of turbines
        name: str
            The constraint name

        """
        super().__init__(
            problem,
            name,
            vnames_int=problem.var_names_int(),
            vnames_float=problem.var_names_float(),
        )
        self.N = N

    def n_components(self):
        """
        Returns the number of components of the
        function.

        Returns
        -------
        int:
            The number of components.

        """
        return 1

    def maximize(self):
        """
        Returns flag for maximization of each component.

        Returns
        -------
        flags: np.array
            Bool array for component maximization,
            shape: (n_components,)

        """
        return [False]

    def calc_individual(self, vars_int, vars_float, problem_results, cmpnts=None):
        """
        Calculate values for a single individual of the
        underlying problem.

        Parameters
        ----------
        vars_int : np.array
            The integer variable values, shape: (n_vars_int,)
        vars_float : np.array
            The float variable values, shape: (n_vars_float,)
        problem_results : Any
            The results of the variable application
            to the problem
        components : list of int, optional
            The selected components or None for all

        Returns
        -------
        values : np.array
            The component values, shape: (n_sel_components,)

        """
        __, valid = problem_results
        N = np.sum(valid, dtype=np.float64)
        return np.maximum(N - self.N, self.N - N)

    def calc_population(self, vars_int, vars_float, problem_results, cmpnts=None):
        """
        Calculate values for all individuals of a population.

        Parameters
        ----------
        vars_int : np.array
            The integer variable values, shape: (n_pop, n_vars_int)
        vars_float : np.array
            The float variable values, shape: (n_pop, n_vars_float)
        problem_results : Any
            The results of the variable application
            to the problem
        components : list of int, optional
            The selected components or None for all

        Returns
        -------
        values : np.array
            The component values, shape: (n_pop, n_sel_components)

        """
        __, valid = problem_results
        N = np.sum(valid, axis=1, dtype=np.float64)[:, None]
        return np.maximum(N - self.N, self.N - N)


class MaxGridSpacing(Objective):
    """
    Maximal grid spacing objective
    for purely geometrical layouts problems.

    :group: opt.problems.layout.geom_layouts.objectives

    """

    def __init__(self, problem, name="max_dxdy"):
        """
        Constructor.

        Parameters
        ----------
        problem: foxes.opt.FarmOptProblem
            The underlying geometrical layout
            optimization problem
        name: str
            The constraint name

        """
        super().__init__(
            problem,
            name,
            vnames_int=problem.var_names_int(),
            vnames_float=problem.var_names_float(),
        )

    def n_components(self):
        """
        Returns the number of components of the
        function.

        Returns
        -------
        int:
            The number of components.

        """
        return 1

    def maximize(self):
        """
        Returns flag for maximization of each component.

        Returns
        -------
        flags: np.array
            Bool array for component maximization,
            shape: (n_components,)

        """
        return [True]

    def calc_individual(self, vars_int, vars_float, problem_results, cmpnts=None):
        """
        Calculate values for a single individual of the
        underlying problem.

        Parameters
        ----------
        vars_int : np.array
            The integer variable values, shape: (n_vars_int,)
        vars_float : np.array
            The float variable values, shape: (n_vars_float,)
        problem_results : Any
            The results of the variable application
            to the problem
        components : list of int, optional
            The selected components or None for all

        Returns
        -------
        values : np.array
            The component values, shape: (n_sel_components,)

        """
        vflt = vars_float.reshape(self.problem.n_grids, 5)
        delta = np.minimum(vflt[:, 2], vflt[:, 3])
        return np.nanmin(delta)

    def calc_population(self, vars_int, vars_float, problem_results, cmpnts=None):
        """
        Calculate values for all individuals of a population.

        Parameters
        ----------
        vars_int : np.array
            The integer variable values, shape: (n_pop, n_vars_int)
        vars_float : np.array
            The float variable values, shape: (n_pop, n_vars_float)
        problem_results : Any
            The results of the variable application
            to the problem
        components : list of int, optional
            The selected components or None for all

        Returns
        -------
        values : np.array
            The component values, shape: (n_pop, n_sel_components)

        """
        n_pop = vars_float.shape[0]
        vflt = vars_float.reshape(n_pop, self.problem.n_grids, 5)
        delta = np.minimum(vflt[:, :, 2], vflt[:, :, 3])
        return np.nanmin(delta, axis=1)[:, None]


class MaxDensity(Objective):
    """
    Maximal turbine density objective
    for purely geometrical layouts problems.

    :group: opt.problems.layout.geom_layouts.objectives

    """

    def __init__(self, problem, dfactor=1, min_dist=None, name="max_density"):
        """
        Constructor.

        Parameters
        ----------
        problem: foxes.opt.FarmOptProblem
            The underlying geometrical layout
            optimization problem
        dfactor: float
            Delta factor for grid spacing
        min_dist: float, optional
            The minimal distance
        name: str
            The constraint name

        """
        super().__init__(
            problem,
            name,
            vnames_int=problem.var_names_int(),
            vnames_float=problem.var_names_float(),
        )
        self.dfactor = dfactor
        self.min_dist = problem.min_dist if min_dist is None else min_dist

    def n_components(self):
        """
        Returns the number of components of the
        function.

        Returns
        -------
        int:
            The number of components.

        """
        return 1

    def maximize(self):
        """
        Returns flag for maximization of each component.

        Returns
        -------
        flags: np.array
            Bool array for component maximization,
            shape: (n_components,)

        """
        return [False]

    def initialize(self, verbosity):
        """
        Initialize the object.

        Parameters
        ----------
        verbosity: int
            The verbosity level, 0 = silent

        """
        super().initialize(verbosity)

        # define regular grid of probe points:
        geom = self.problem.boundary
        pmin = geom.p_min()
        pmax = geom.p_max()
        detlta = self.min_dist / self.dfactor
        self._probes = np.stack(
            np.meshgrid(
                np.arange(pmin[0] - detlta, pmax[0] + 2 * detlta, detlta),
                np.arange(pmin[1] - detlta, pmax[1] + 2 * detlta, detlta),
                indexing="ij",
            ),
            axis=-1,
        )
        nx, ny = self._probes.shape[:2]
        n = nx * ny
        self._probes = self._probes.reshape(n, 2)

        # reduce to points within geometry:
        valid = geom.points_inside(self._probes)
        self._probes = self._probes[valid]

    def calc_individual(self, vars_int, vars_float, problem_results, cmpnts=None):
        """
        Calculate values for a single individual of the
        underlying problem.

        Parameters
        ----------
        vars_int : np.array
            The integer variable values, shape: (n_vars_int,)
        vars_float : np.array
            The float variable values, shape: (n_vars_float,)
        problem_results : Any
            The results of the variable application
            to the problem
        components : list of int, optional
            The selected components or None for all

        Returns
        -------
        values : np.array
            The component values, shape: (n_sel_components,)

        """
        xy, valid = problem_results
        xy = xy[valid]
        dists = cdist(self._probes, xy)
        return np.nanmax(np.nanmin(dists, axis=1))

    def calc_population(self, vars_int, vars_float, problem_results, cmpnts=None):
        """
        Calculate values for all individuals of a population.

        Parameters
        ----------
        vars_int : np.array
            The integer variable values, shape: (n_pop, n_vars_int)
        vars_float : np.array
            The float variable values, shape: (n_pop, n_vars_float)
        problem_results : Any
            The results of the variable application
            to the problem
        components : list of int, optional
            The selected components or None for all

        Returns
        -------
        values : np.array
            The component values, shape: (n_pop, n_sel_components)

        """
        n_pop = vars_float.shape[0]
        xy, valid = problem_results
        out = np.full(n_pop, 1e20, dtype=FC.DTYPE)
        for pi in range(n_pop):
            if np.any(valid[pi]):
                hxy = xy[pi][valid[pi]]
                dists = cdist(self._probes, hxy)
                out[pi] = np.nanmax(np.nanmin(dists, axis=1))
        return out[:, None]


class MeMiMaDist(Objective):
    """
    Mean-min-max distance objective
    for purely geometrical layouts problems.

    :group: opt.problems.layout.geom_layouts.objectives

    """

    def __init__(self, problem, scale=500.0, c1=1, c2=1, c3=1, name="MiMaMean"):
        """
        Constructor.

        Parameters
        ----------
        problem: foxes.opt.FarmOptProblem
            The underlying geometrical layout
            optimization problem
        scale: float
            The distance scale
        c1: float
            Parameter for mean weighting
        c2: float
            Parameter for max diff weighting
        c3: float
            Parameter for min diff weighting
        name: str
            The constraint name

        """
        super().__init__(
            problem,
            name,
            vnames_int=problem.var_names_int(),
            vnames_float=problem.var_names_float(),
        )
        self.scale = scale
        self.c1 = c1
        self.c2 = c2
        self.c3 = c3

    def n_components(self):
        """
        Returns the number of components of the
        function.

        Returns
        -------
        int:
            The number of components.

        """
        return 1

    def maximize(self):
        """
        Returns flag for maximization of each component.

        Returns
        -------
        flags: np.array
            Bool array for component maximization,
            shape: (n_components,)

        """
        return [True]

    def calc_individual(self, vars_int, vars_float, problem_results, cmpnts=None):
        """
        Calculate values for a single individual of the
        underlying problem.

        Parameters
        ----------
        vars_int : np.array
            The integer variable values, shape: (n_vars_int,)
        vars_float : np.array
            The float variable values, shape: (n_vars_float,)
        problem_results : Any
            The results of the variable application
            to the problem
        components : list of int, optional
            The selected components or None for all

        Returns
        -------
        values : np.array
            The component values, shape: (n_sel_components,)

        """
        xy, valid = problem_results
        # xy = xy[valid]

        dists = cdist(xy, xy)
        np.fill_diagonal(dists, np.inf)
        dists = np.min(dists, axis=1) / self.scale / len(xy)

        mean = np.average(dists)
        mi = np.min(dists)
        ma = np.max(dists)
        return np.atleast_1d(
            self.c1 * mean**2 - self.c2 * (mean - mi) ** 2 - self.c3 * (mean - ma) ** 2
        )

    def calc_population(self, vars_int, vars_float, problem_results, cmpnts=None):
        """
        Calculate values for all individuals of a population.

        Parameters
        ----------
        vars_int : np.array
            The integer variable values, shape: (n_pop, n_vars_int)
        vars_float : np.array
            The float variable values, shape: (n_pop, n_vars_float)
        problem_results : Any
            The results of the variable application
            to the problem
        components : list of int, optional
            The selected components or None for all

        Returns
        -------
        values : np.array
            The component values, shape: (n_pop, n_sel_components)

        """
        xy, valid = problem_results
        n_pop, n_xy = xy.shape[:2]

        out = np.zeros((n_pop, 1), dtype=FC.DTYPE)
        for pi in range(n_pop):
            hxy = xy[pi]  # , valid[pi]]

            dists = cdist(hxy, hxy)
            np.fill_diagonal(dists, np.inf)
            dists = np.min(dists, axis=1) / self.scale / n_xy

            mean = np.average(dists)
            mi = np.min(dists)
            ma = np.max(dists)
            out[pi, 0] = (
                self.c1 * mean**2
                - self.c2 * (mean - mi) ** 2
                - self.c3 * (mean - ma) ** 2
            )

        return out



import numpy as np

from foxes.core import WakeFrame
from foxes.utils import wd2uv
import foxes.variables as FV
import foxes.constants as FC


class RotorWD(WakeFrame):
    """
    Align the first axis for each rotor with the
    local normalized wind direction.

    Attributes
    ----------
    var_wd: str
        The wind direction variable

    :group: models.wake_frames

    """

    def __init__(self, var_wd=FV.WD):
        """
        Constructor.

        Parameters
        ----------
        var_wd: str
            The wind direction variable

        """
        super().__init__()
        self.var_wd = var_wd

    def calc_order(self, algo, mdata, fdata):
        """ "
        Calculates the order of turbine evaluation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data

        Returns
        -------
        order: numpy.ndarray
            The turbine order, shape: (n_states, n_turbines)

        """
        n = np.mean(wd2uv(fdata[self.var_wd], axis=1), axis=-1)
        xy = fdata[FV.TXYH][:, :, :2]
        order = np.argsort(np.einsum("std,sd->st", xy, n), axis=-1)

        return order

    def get_wake_coos(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
    ):
        """
        Calculate wake coordinates of rotor points.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwnd order

        Returns
        -------
        wake_coos: numpy.ndarray
            The wake frame coordinates of the evaluation
            points, shape: (n_states, n_targets, n_tpoints, 3)

        """
        n_states = tdata.n_states
        targets = tdata[FC.TARGETS]

        xyz = fdata[FV.TXYH][:, downwind_index]
        delta = targets - xyz[:, None, None, :]
        del xyz

        wd = fdata[self.var_wd][:, downwind_index]

        nax = np.zeros((n_states, 3, 3), dtype=FC.DTYPE)
        n = nax[:, 0, :2]
        n[:] = wd2uv(wd, axis=-1)
        m = nax[:, 1, :2]
        m[:] = np.stack([-n[:, 1], n[:, 0]], axis=-1)
        nax[:, 2, 2] = 1
        del wd

        coos = np.einsum("stpd,sad->stpa", delta, nax)

        return coos

    def get_centreline_points(self, algo, mdata, fdata, downwind_index, x):
        """
        Gets the points along the centreline for given
        values of x.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        downwind_index: int
            The index in the downwind order
        x: numpy.ndarray
            The wake frame x coordinates, shape: (n_states, n_points)

        Returns
        -------
        points: numpy.ndarray
            The centreline points, shape: (n_states, n_points, 3)

        """
        wd = fdata[self.var_wd][:, downwind_index]
        n = np.append(wd2uv(wd, axis=-1), np.zeros_like(wd)[:, None], axis=-1)

        xyz = fdata[FV.TXYH][:, downwind_index]
        return xyz[:, None, :] + x[:, :, None] * n[:, None, :]



import pandas as pd
from xarray import Dataset

import foxes.constants as FC
from foxes.utils import write_nc

from .output import Output


class ResultsWriter(Output):
    """
    Writes results farm data to file.

    Attributes
    ----------
    data: pandas.DataFrame
        The farm results

    :group: output

    """

    def __init__(self, farm_results=None, data=None, **kwargs):
        """
        Constructor.

        Parameters
        ----------
        farm_results: xarray.Dataset, optional
            The farm results, if data is None
        data: pandas.DataFrame, optional
            The data, if farm_results is None
        kwargs: dict, optional
            Additional parameters for the base class

        """
        super().__init__(**kwargs)

        if farm_results is not None and data is None:
            self.data = farm_results.to_dataframe().reset_index()
            self.data[FC.TNAME] = farm_results[FC.TNAME].to_numpy()[
                self.data[FC.TURBINE]
            ]
            self.data.set_index([FC.STATE, FC.TURBINE], inplace=True)
        elif farm_results is None and data is not None:
            self.data = data
        else:
            raise KeyError(
                f"ResultsWriter: Either give 'farm_results' or 'data' arguments"
            )

    def _get_data_vars(self, variables):
        """Helper function for variable gathering"""
        data = self.data
        if variables is None:
            pass
        elif isinstance(variables, dict):
            inds = {
                s: variables.pop(s) for s in self.data.index.names if s in variables
            }
            if len(variables):
                data = data[list(variables.keys())].rename(variables, axis=1)
            if len(inds):
                for s, ns in inds.items():
                    data = data.rename_axis(index={s: ns})
        else:
            data = self.data[list(variables)]

        return data, data.columns.tolist()

    def write_csv(
        self,
        file_path,
        variables=None,
        turbine_names=False,
        state_turbine_table=False,
        verbosity=1,
        **kwargs,
    ):
        """
        Writes a csv file

        Parameters
        ----------
        file_path: str
            Path the the csv file
        variables: dict or list of str, optional
            The variables to be written. If a dict, then
            the keys are the foxes variables and the values
            the column names. If None, then all data will be
            written.
        turbine_names: bool
            Use turbine names instead of turbine indices
        state_turbine_table: bool
            Flag for writing a single variable into turbine columns
            for state rows
        verbosity: int
            The verbosity level, 0 = silent
        kwargs: dict, optional
            Additional parameters for Output.write()

        """
        if verbosity:
            print(f"ResultsWriter: Writing file '{file_path}'")

        data, variables = self._get_data_vars(variables)
        data.reset_index(inplace=True)
        fc2v = kwargs.pop("format_col2var", {})
        tix = FC.TNAME if turbine_names else FC.TURBINE

        if state_turbine_table:
            if len(variables) != 1:
                raise ValueError(
                    f"state_turbine_table can only be written for a single variable, got {variables}"
                )
            v = variables[0]
            cnames = {
                t: f"{v}_T{t:04d}" if tix == FC.TURBINE else f"{v}_{t}"
                for t in data[tix]
            }
            for ti, (t, g) in enumerate(
                data.reset_index().set_index(FC.STATE).groupby(tix)
            ):
                if ti == 0:
                    odata = pd.DataFrame(
                        index=g.index.to_numpy(), columns=list(cnames.values())
                    )
                    odata.index.name = g.index.name
                cname = cnames[t]
                odata[cname] = g[v].to_numpy().copy()
                fc2v[cname] = v
            data = odata
        else:
            data.set_index([FC.STATE, tix], inplace=True)

        fpath = self.get_fpath(file_path)
        super().write(fpath, data, format_col2var=fc2v, **kwargs)

    def write_nc(
        self,
        file_path,
        variables=None,
        turbine_names=False,
        verbosity=1,
        **kwargs,
    ):
        """
        Writes a netCDF file

        Parameters
        ----------
        file_path: str
            Path the the csv file
        variables: dict or list of str, optional
            The variables to be written. If a dict, then
            the keys are the foxes variables and the values
            the column names. If None, then all data will be
            written.
        turbine_names: bool
            Use turbine names instead of turbine indices
        verbosity: int
            The verbosity level, 0 = silent
        kwargs: dict, optional
            Additional parameters for write_nc()

        """
        data, variables = self._get_data_vars(variables)
        states = data.index.get_level_values(0).unique()
        n_states = len(states)
        n_turbines = len(data.index.get_level_values(1).unique())

        crds = {FC.STATE: states}
        if turbine_names:
            idx = pd.IndexSlice
            crds[FC.TURBINE] = self.data.loc[idx[0, :], FC.TNAME].to_numpy()

        ds = Dataset(
            coords=crds,
            data_vars={
                v: (
                    (FC.STATE, FC.TURBINE),
                    data[v].to_numpy().reshape(n_states, n_turbines),
                )
                for v in variables
            },
        )

        fpath = self.get_fpath(file_path)
        write_nc(ds, fpath, verbosity=verbosity, **kwargs)



import numpy as np
from abc import abstractmethod

import foxes.variables as FV
import foxes.constants as FC
from .farm_data_model import FarmDataModel
from foxes.utils import wd2uv, uv2wd, all_subclasses

from .data import TData


class RotorModel(FarmDataModel):
    """
    Abstract base class of rotor models.

    Rotor models calculate ambient farm data from
    states, and provide rotor points and weights
    for the calculation of rotor effective quantities.

    Attributes
    ----------
    calc_vars: list of str
        The variables that are calculated by the model
        (Their ambients are added automatically)

    :group: core

    """

    def __init__(self, calc_vars):
        """
        Constructor.

        Parameters
        ----------
        calc_vars: list of str
            The variables that are calculated by the model
            (Their ambients are added automatically)

        """
        super().__init__()
        self.calc_vars = calc_vars

        self.RPOINTS = self.var("rpoints")
        self.RWEIGHTS = self.var("rweights")
        self.AMBRES = self.var("amb_res")

    def output_farm_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return list(
            set(
                self.calc_vars
                + [FV.var2amb[v] for v in self.calc_vars if v in FV.var2amb]
            )
        )

    @abstractmethod
    def n_rotor_points(self):
        """
        The number of rotor points

        Returns
        -------
        n_rpoints: int
            The number of rotor points

        """
        pass

    @abstractmethod
    def rotor_point_weights(self):
        """
        The weights of the rotor points

        Returns
        -------
        weights: numpy.ndarray
            The weights of the rotor points,
            add to one, shape: (n_rpoints,)

        """
        pass

    @abstractmethod
    def design_points(self):
        """
        The rotor model design points.

        Design points are formulated in rotor plane
        (x,y,z)-coordinates in rotor frame, such that
        - (0,0,0) is the centre point,
        - (1,0,0) is the point radius * n_rotor_axis
        - (0,1,0) is the point radius * n_rotor_side
        - (0,0,1) is the point radius * n_rotor_up

        Returns
        -------
        dpoints: numpy.ndarray
            The design points, shape: (n_points, 3)

        """
        pass

    def get_rotor_points(self, algo, mdata, fdata):
        """
        Calculates rotor points from design points.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data

        Returns
        -------
        points: numpy.ndarray
            The rotor points, shape:
            (n_states, n_turbines, n_rpoints, 3)

        """

        n_states = mdata.n_states
        n_points = self.n_rotor_points()
        n_turbines = algo.n_turbines
        dpoints = self.design_points()
        D = fdata[FV.D]

        rax = np.zeros((n_states, n_turbines, 3, 3), dtype=FC.DTYPE)
        n = rax[:, :, 0, 0:2]
        m = rax[:, :, 1, 0:2]
        n[:] = wd2uv(fdata[FV.YAW], axis=-1)
        m[:] = np.stack([-n[:, :, 1], n[:, :, 0]], axis=-1)
        rax[:, :, 2, 2] = 1

        points = np.zeros((n_states, n_turbines, n_points, 3), dtype=FC.DTYPE)
        points[:] = fdata[FV.TXYH][:, :, None, :]
        points[:] += (
            0.5 * D[:, :, None, None] * np.einsum("stad,pa->stpd", rax, dpoints)
        )

        return points

    def _set_res(self, fdata, v, res, downwind_index):
        """
        Helper function for results setting
        """
        if downwind_index is None:
            fdata[v] = res.copy()
        elif res.shape[1] == 1:
            fdata[v][:, downwind_index] = res[:, 0]
        else:
            raise ValueError(
                f"Rotor model '{self.name}': downwind_index is not None, but results shape for '{v}' has more than one turbine, {res.shape}"
            )

    def eval_rpoint_results(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        weights,
        downwind_index=None,
        copy_to_ambient=False,
    ):
        """
        Evaluate rotor point results.

        This function modifies `fdata`, either
        for all turbines or one turbine per state,
        depending on parameter `states_turbine`. In
        the latter case, the turbine dimension of the
        `rpoint_results` is expected to have size one.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        weights: numpy.ndarray
            The rotor point weights, shape: (n_rpoints,)
        downwind_index: int, optional
            The index in the downwind order
        copy_to_ambient: bool
            If `True`, the fdata results are copied to ambient
            variables after calculation

        """
        uvp = None
        uv = None
        if (
            FV.WS in self.calc_vars
            or FV.WD in self.calc_vars
            or FV.YAW in self.calc_vars
            or FV.REWS in self.calc_vars
            or FV.REWS2 in self.calc_vars
            or FV.REWS3 in self.calc_vars
        ):
            wd = tdata[FV.WD]
            ws = tdata[FV.WS]
            uvp = wd2uv(wd, ws, axis=-1)
            uv = np.einsum("stpd,p->std", uvp, weights)

        wd = None
        vdone = []
        for v in self.calc_vars:
            if v == FV.WD or v == FV.YAW:
                if wd is None:
                    wd = uv2wd(uv, axis=-1)
                self._set_res(fdata, v, wd, downwind_index)
                vdone.append(v)

            elif v == FV.WS:
                ws = np.linalg.norm(uv, axis=-1)
                self._set_res(fdata, v, ws, downwind_index)
                del ws
                vdone.append(v)
        del uv, wd

        if (
            FV.REWS in self.calc_vars
            or FV.REWS2 in self.calc_vars
            or FV.REWS3 in self.calc_vars
        ):
            if downwind_index is None:
                yaw = fdata[FV.YAW].copy()
            else:
                yaw = fdata[FV.YAW][:, downwind_index, None]
            nax = wd2uv(yaw, axis=-1)
            wsp = np.einsum("stpd,std->stp", uvp, nax)

            for v in self.calc_vars:
                if v == FV.REWS:
                    rews = np.maximum(np.einsum("stp,p->st", wsp, weights), 0.0)
                    self._set_res(fdata, v, rews, downwind_index)
                    del rews
                    vdone.append(v)

                elif v == FV.REWS2:
                    # For highly inhomogeneous wind fields
                    # and multiple rotor points some of the uv
                    # vectors may have negative projections onto the
                    # turbine axis direction:
                    if uvp.shape[2] > 1:
                        rews2 = np.sqrt(
                            np.maximum(
                                np.einsum("stp,p->st", np.sign(wsp) * wsp**2, weights),
                                0.0,
                            )
                        )
                    else:
                        rews2 = np.sqrt(np.einsum("stp,p->st", wsp**2, weights))
                    self._set_res(fdata, v, rews2, downwind_index)
                    del rews2
                    vdone.append(v)

                elif v == FV.REWS3:
                    # For highly inhomogeneous wind fields
                    # and multiple rotor points some of the uv
                    # vectors may have negative projections onto the
                    # turbine axis direction:
                    if uvp.shape[2] > 1:
                        rews3 = np.maximum(
                            np.einsum("stp,p->st", wsp**3, weights), 0.0
                        ) ** (1.0 / 3.0)
                    else:
                        rews3 = (np.einsum("stp,p->st", wsp**3, weights)) ** (1.0 / 3.0)
                    self._set_res(fdata, v, rews3, downwind_index)
                    del rews3
                    vdone.append(v)

            del wsp
        del uvp

        for v in self.calc_vars:
            if v not in vdone:
                res = np.einsum("stp,p->st", tdata[v], weights)
                self._set_res(fdata, v, res, downwind_index)
            if copy_to_ambient and v in FV.var2amb:
                fdata[FV.var2amb[v]] = fdata[v].copy()

    def calculate(
        self,
        algo,
        mdata,
        fdata,
        rpoints=None,
        weights=None,
        store_rpoints=False,
        store_rweights=False,
        store_amb_res=False,
        downwind_index=None,
    ):
        """
        Calculate ambient rotor effective results.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        rpoints: numpy.ndarray, optional
            The rotor points, or None for automatic for
            this rotor. Shape: (n_states, n_turbines, n_rpoints, 3)
        weights: numpy.ndarray, optional
            The rotor point weights, or None for automatic
            for this rotor. Shape: (n_rpoints,)
        store_rpoints: bool, optional
            Switch for storing rotor points to mdata
        store_rweights: bool, optional
            Switch for storing rotor point weights to mdata
        store_amb_res: bool, optional
            Switch for storing ambient rotor point reults as they
            come from the states to mdata
        downwind_index: int, optional
            Only compute for index in the downwind order

        Returns
        -------
        results: dict
            results dict. Keys: Variable name str. Values:
            numpy.ndarray with results, shape: (n_states, n_turbines)

        """

        if rpoints is None:
            rpoints = mdata.get(self.RPOINTS, self.get_rotor_points(algo, mdata, fdata))
        if store_rpoints:
            mdata[self.RPOINTS] = rpoints
            mdata.dims[self.RPOINTS] = (FC.STATE, FC.TURBINE, FC.TPOINT, FC.XYH)
            self.data_to_store(self.RPOINTS, algo, mdata)

        if downwind_index is not None:
            rpoints = rpoints[:, downwind_index, None]

        if weights is None:
            weights = mdata.get(FC.TWEIGHTS, self.rotor_point_weights())
        if store_rweights:
            mdata[self.RWEIGHTS] = weights
            mdata.dims[self.RWEIGHTS] = (FC.TPOINT,)
            self.data_to_store(self.RWEIGHTS, algo, mdata)

        tdata = TData.from_tpoints(rpoints, weights)
        svars = algo.states.output_point_vars(algo)
        for v in svars:
            tdata.add(
                v,
                data=np.full_like(rpoints[..., 0], np.nan),
                dims=(FC.STATE, FC.TARGET, FC.TPOINT),
            )

        sres = algo.states.calculate(algo, mdata, fdata, tdata)
        tdata.update(sres)

        if store_amb_res:
            mdata[self.AMBRES] = sres.copy()
            self.data_to_store(self.AMBRES, algo, mdata)

        self.eval_rpoint_results(
            algo,
            mdata,
            fdata,
            tdata,
            weights,
            downwind_index,
            copy_to_ambient=True,
        )

        return {v: fdata[v] for v in self.output_farm_vars(algo)}

    @classmethod
    def new(cls, rmodel_type, *args, **kwargs):
        """
        Run-time rotor model factory.

        Parameters
        ----------
        rmodel_type: str
            The selected derived class name
        args: tuple, optional
            Additional parameters for constructor
        kwargs: dict, optional
            Additional parameters for constructor

        """

        if rmodel_type is None:
            return None

        allc = all_subclasses(cls)
        found = rmodel_type in [scls.__name__ for scls in allc]

        if found:
            for scls in allc:
                if scls.__name__ == rmodel_type:
                    return scls(*args, **kwargs)

        else:
            estr = "Rotor model type '{}' is not defined, available types are \n {}".format(
                rmodel_type, sorted([i.__name__ for i in allc])
            )
            raise KeyError(estr)



import numpy as np
import pandas as pd

from .PCt_file import PCtFile
from foxes.data import parse_Pct_file_name
from foxes.utils import PandasFileHelper
import foxes.constants as FC


class CpCtFile(PCtFile):
    """
    Calculate power and ct by interpolating
    from cp-ct-curve data file (or pandas DataFrame).

    :group: models.turbine_types

    """

    def __init__(
        self,
        data_source,
        col_ws="ws",
        col_cp="cp",
        rho=1.225,
        pd_file_read_pars={},
        **parameters,
    ):
        """
        Constructor.

        Parameters
        ----------
        data_source: str or pandas.DataFrame
            The file path, static name, or data
        col_ws: str
            The wind speed column
        col_cp: str
            The cp column
        rho: float
            The air density for the curves
        pd_file_read_pars: dict
            Parameters for pandas file reading
        paramerers: dict, optional
            Additional parameters for PCtFile class

        """
        if not isinstance(data_source, pd.DataFrame):
            pars = parse_Pct_file_name(data_source)
            pars.update(parameters)
            data = PandasFileHelper.read_file(data_source, **pd_file_read_pars)
        else:
            data = data_source
            pars = parameters

        D = pars["D"]
        A = np.pi * (D / 2) ** 2
        ws = data[col_ws].to_numpy()
        cp = data[col_cp].to_numpy()
        P_unit = pars.pop("P_unit", FC.kW)

        ws_delta = 0.0001
        ws_min = np.min(ws)
        ws_max = np.max(ws)
        N = int((ws_max - ws_min) / ws_delta)

        data_P = pd.DataFrame(index=range(N), dtype=FC.DTYPE)
        data_P["ws"] = np.linspace(ws_min, ws_max, N, endpoint=True)
        data_P["cp"] = np.interp(data_P["ws"], ws, cp, left=0, right=0)
        data_P["P"] = (
            0.5 * rho * A * data_P["cp"] * data_P["ws"] ** 3 / FC.P_UNITS[P_unit]
        )

        super().__init__(data_P, col_ws="ws", col_P="P", rho=rho, P_unit=P_unit, **pars)



import numpy as np


def wd2wdvec(wd, ws=1.0, axis=-1):
    """
    Calculate wind direction vectors from wind directions
    in degrees.

    Parameters
    ----------
    wd: numpy.ndarray
        Wind direction array (any shape)
    ws: float or numpy.ndarray
        The wind speed. Has to broadcast against wd.
    axis: int
        Location where to insert the (x, y) dimension
        into the shape of wd

    Returns
    -------
    wdvec: numpy.ndarray
        The wind direction vectors

    :group: utils

    """

    wdr = wd * np.pi / 180.0
    n = np.stack([np.sin(wdr), np.cos(wdr)], axis=axis)

    if np.isscalar(ws):
        return ws * n

    return np.expand_dims(ws, axis) * n


def wd2uv(wd, ws=1.0, axis=-1):
    """
    Calculate wind vectors from wind directions
    in degrees.

    Parameters
    ----------
    wd: numpy.ndarray
        Wind direction array (any shape)
    ws: float or numpy.ndarray
        The wind speed. Has to broadcast against wd.
    axis: int
        Axis location where to insert the (u, v) components
        into the shape of wd

    Returns
    -------
    uv: numpy.ndarray
        The wind vectors

    :group: utils

    """
    return -wd2wdvec(wd, ws, axis)


def uv2wd(uv, axis=-1):
    """
    Calculate wind direction from wind vectors.

    Parameters
    ----------
    uv: numpy.ndarray
        The wind vectors, any shape
    axis: int
        The axis which corresponds to (u, v) components

    Returns
    -------
    wd: numpy.ndarray
        The wind direction array

    :group: utils

    """

    if axis == -1:
        u = uv[..., 0]
        v = uv[..., 1]
    else:
        s = tuple(0 if a == axis else slice(None) for a in range(len(uv.shape)))
        u = uv[s]
        s = tuple(1 if a == axis else slice(None) for a in range(len(uv.shape)))
        v = uv[s]

    return np.mod(180 + np.rad2deg(np.arctan2(u, v)), 360)


def wdvec2wd(wdvec, axis=-1):
    """
    Calculate wind direction from wind direction vectors.

    Parameters
    ----------
    wdvec: numpy.ndarray
        The wind direction vectors, any shape
    axis: int
        The axis which corresponds to (x, y) components

    Returns
    -------
    wd: numpy.ndarray
        The wind direction array

    :group: utils

    """
    return uv2wd(-wdvec, axis)


def delta_wd(wd_a, wd_b):
    """
    Calculates wd_b - wd_a.

    Parameters
    ----------
    wd_a: numpy.ndarray
        Array of wind directions.
        Shape: any shape
    wd_b: numpy.ndarray
        Array of wind directions.
        Shape: same as wd_a

    Returns
    -------
    numpy.ndarray :
        Array of wind direction deltas.
        Shape: same as wd_a, wd_b

    :group: utils

    """
    out = wd_b - wd_a

    out[out < -180.0] += 360.0
    out[out > 180.0] -= 360.0

    return out



import numpy as np
import pandas as pd

from foxes.opt.core import FarmVarsProblem
from foxes.models.turbine_models import SetFarmVars
import foxes.constants as FC


class OptFarmVars(FarmVarsProblem):
    """
    Optimize a selection of farm variables.

    :group: opt.problems

    """

    def __init__(self, *args, **kwargs):
        """
        Constructor.

        Parameters
        ----------
        args: tuple, optional
            Arguments for `FarmVarsProblem`
        kwargs: dict, optional
            Keyword arguments for `FarmVarsProblem`

        """
        super().__init__(*args, **kwargs)
        self._vars = None

    def add_var(
        self,
        name,
        typ,
        init,
        min,
        max,
        level="uniform",
        sel=None,
        pre_rotor=False,
        model_key=None,
    ):
        """
        Add a variable.

        Parameters
        ----------
        name: str
            The foxes farm variable name
        typ: type
            The variable type, either float or int
        init: float or int
            The initial value
        min: float or int
            The min value
        max: float or int
            The max value
        level: str
            Choices: uniform, state, turbine, state-turbine
        sel: numpy.ndarray, optional
            States/turbines/state-turbine selection,
            depending on the level
        pre_rotor: bool
            Apply this variable before rotor model
        model_key: str, optional
            Creates sub-model which can then be placed in the
            turbine model list. Repeated keys are added to the
            same turbine model

        """
        if typ is not float and typ is not int:
            raise TypeError(
                f"Problem '{self.name}': Expecting float or int, got type '{type(typ).__name__}'"
            )

        mname = self.name if model_key is None else model_key
        if mname in self.algo.mbook.turbine_models:
            m = self.algo.mbook.turbine_models[mname]
            if not isinstance(m, SetFarmVars):
                raise KeyError(
                    f"Problem '{self.name}': Turbine model entry '{mname}' already exists in model book, and is not of type SetFarmVars"
                )
            elif m.pre_rotor != pre_rotor:
                raise ValueError(
                    f"Problem '{self.name}': Turbine model entry '{mname}' exists in model book, and disagrees on pre_rotor = {pre_rotor}"
                )
        else:
            self.algo.mbook.turbine_models[mname] = SetFarmVars(pre_rotor=pre_rotor)

        if self._vars is None:
            i0 = 0
            i0i = 0
            i0f = 0
        else:
            if name in self._vars["var"].tolist():
                raise ValueError(
                    f"Problem '{self.name}': Attempt to add variable '{name}' twice"
                )
            i0 = len(self._vars.index)
            grps = self._vars.groupby("type")
            i0i = len(grps.get_group("int").index) if "int" in grps.groups.keys() else 0
            i0f = (
                len(grps.get_group("float").index)
                if "float" in grps.groups.keys()
                else 0
            )
            del grps

        if level == "uniform":
            hdata = pd.DataFrame(index=[i0])
            hdata.loc[i0, "name"] = name
            hdata.loc[i0, "var"] = name
            hdata.loc[i0, "type"] = "int" if typ is int else "float"
            hdata.loc[i0, "index"] = i0i if typ is int else i0f
            hdata.loc[i0, "level"] = level
            hdata.loc[i0, "state"] = -1
            hdata.loc[i0, "turbine"] = -1
            hdata.loc[i0, "sel_turbine"] = -1
            hdata.loc[i0, "init"] = np.array([init], dtype=FC.DTYPE)
            hdata.loc[i0, "min"] = np.array([min], dtype=FC.DTYPE)
            hdata.loc[i0, "max"] = np.array([max], dtype=FC.DTYPE)
            hdata.loc[i0, "pre_rotor"] = pre_rotor
            hdata.loc[i0, "model_key"] = mname

        elif level == "state":
            if not self.algo.initialized:
                self.algo.initialize()

            states = np.arange(self.algo.n_states)
            if sel is not None:
                states = states[sel]
            inds = i0 + np.arange(len(states))
            tinds = inds - i0 + i0i if typ is int else inds - i0 + i0f

            hdata = pd.DataFrame(index=inds)
            hdata.loc[inds, "name"] = [f"{name}_{i:05d}" for i in range(len(states))]
            hdata.loc[inds, "var"] = name
            hdata.loc[inds, "type"] = "int" if typ is int else "float"
            hdata.loc[inds, "index"] = tinds
            hdata.loc[inds, "level"] = level
            hdata.loc[inds, "state"] = states
            hdata.loc[inds, "turbine"] = -1
            hdata.loc[inds, "sel_turbine"] = -1

            for c, d in [("init", init), ("min", min), ("max", max)]:
                data = np.full(len(inds), np.nan, dtype=FC.DTYPE)
                data[:] = d
                hdata.loc[inds, c] = data

            hdata.loc[inds, "pre_rotor"] = pre_rotor
            hdata.loc[inds, "model_key"] = mname

        elif level == "turbine":
            if sel is None:
                turbines = self.sel_turbines
            else:
                turbines = np.arange(self.algo.n_turbines)[sel]
            inds = i0 + np.arange(len(turbines))
            tinds = inds - i0 + i0i if typ is int else inds - i0 + i0f

            hdata = pd.DataFrame(index=inds)
            hdata.loc[inds, "name"] = [f"{name}_{i:04d}" for i in range(len(turbines))]
            hdata.loc[inds, "var"] = name
            hdata.loc[inds, "type"] = "int" if typ is int else "float"
            hdata.loc[inds, "index"] = tinds
            hdata.loc[inds, "level"] = level
            hdata.loc[inds, "state"] = -1
            hdata.loc[inds, "turbine"] = turbines
            hdata.loc[inds, "sel_turbine"] = [
                self.sel_turbines.index(ti) for ti in turbines
            ]

            for c, d in [("init", init), ("min", min), ("max", max)]:
                data = np.full(len(inds), np.nan, dtype=FC.DTYPE)
                data[:] = d
                hdata.loc[inds, c] = data

            hdata.loc[inds, "pre_rotor"] = pre_rotor
            hdata.loc[inds, "model_key"] = mname

        elif level == "state-turbine":
            if not self.algo.initialized:
                self.algo.initialize()

            n_states = self.algo.n_states
            n_turbines = self.algo.n_turbines
            if sel is None:
                sel = np.zeros((n_states, n_turbines), dtype=bool)
                sel[:, self.sel_turbines] = True
            else:
                sel = np.array(sel, dtype=bool)
            st = np.arange(n_states * n_turbines).reshape(n_states, n_turbines)[sel]
            whr = np.where(sel)
            n_inds = len(st)
            inds = i0 + np.arange(n_inds)
            tinds = inds - i0 + i0i if typ is int else inds - i0 + i0f

            hdata = pd.DataFrame(index=inds)
            hdata.loc[inds, "name"] = [
                f"{name}_{whr[0][i]:05d}_{whr[1][i]:04d}" for i in range(len(st))
            ]
            hdata.loc[inds, "var"] = name
            hdata.loc[inds, "type"] = "int" if typ is int else "float"
            hdata.loc[inds, "index"] = tinds
            hdata.loc[inds, "level"] = level
            hdata.loc[inds, "state"] = whr[0]
            hdata.loc[inds, "turbine"] = whr[1]
            hdata.loc[inds, "sel_turbine"] = [
                self.sel_turbines.index(ti) for ti in whr[1]
            ]

            for c, d in [("init", init), ("min", min), ("max", max)]:
                data = np.full(n_inds, np.nan, dtype=FC.DTYPE)
                if isinstance(d, np.ndarray) and len(d.shape) > 1:
                    data[:] = d[sel]
                else:
                    data[:] = d
                hdata.loc[inds, c] = data

            hdata.loc[inds, "pre_rotor"] = pre_rotor
            hdata.loc[inds, "model_key"] = mname

        else:
            raise ValueError(
                f"Problem '{self.name}': Unknown level '{level}'. Choices: uniform, state, turbine, state-turbine"
            )

        if self._vars is None:
            self._vars = hdata
        else:
            self._vars = pd.concat([self._vars, hdata], axis=0)

        icols = ["index", "state", "turbine", "sel_turbine"]
        for c in icols:
            self._vars[c] = self._vars[c].astype(FC.ITYPE)

    def initialize(self, verbosity=1, **kwargs):
        """
        Initialize the object.

        Parameters
        ----------
        verbosity: int
            The verbosity level, 0 = silent
        kwargs: dict, optional
            Additional parameters for super class init

        """
        if self._vars is None:
            raise ValueError(
                f"Problem '{self.name}': No variables added for optimization."
            )

        if verbosity > 0:
            print(f"Problem '{self.name}': Optimization variable list")
            print()
            print(self._vars)
            print()

        prev = {}
        postv = {}
        for (mname, pre), g in self._vars.groupby(["model_key", "pre_rotor"]):
            if (pre and mname in postv) or (not pre and mname in prev):
                raise ValueError(
                    f"Problem '{self.name}': Model '{mname}' reveived both pre_rotor and non-pre_rotor variables"
                )
            tg = prev if pre else postv
            if mname not in tg:
                tg[mname] = set(g["var"].tolist())
            else:
                tg[mname] = tg[mname].update(g["var"].tolist())

        super().initialize(
            pre_rotor_vars={mname: list(vrs) for mname, vrs in prev.items()},
            post_rotor_vars={mname: list(vrs) for mname, vrs in postv.items()},
            verbosity=verbosity,
            **kwargs,
        )

    def var_names_int(self):
        """
        The names of int variables.

        Returns
        -------
        names: list of str
            The names of the int variables

        """
        if self._vars is None:
            raise ValueError(
                f"Problem '{self.name}': No variables added for optimization."
            )

        grps = self._vars.groupby("type")
        if "int" not in grps.groups.keys():
            return []
        else:
            return grps.get_group("int")["name"].tolist()

    def initial_values_int(self):
        """
        The initial values of the int variables.

        Returns
        -------
        values: numpy.ndarray
            Initial int values, shape: (n_vars_int,)

        """
        if self._vars is None:
            raise ValueError(
                f"Problem '{self.name}': No variables added for optimization."
            )

        grps = self._vars.groupby("type")
        if "int" not in grps.groups.keys():
            return []
        else:
            return grps.get_group("int")["init"].to_numpy(FC.ITYPE)

    def min_values_int(self):
        """
        The minimal values of the integer variables.

        Use -self.INT_INF for unbounded.

        Returns
        -------
        values: numpy.ndarray
            Minimal int values, shape: (n_vars_int,)

        """
        if self._vars is None:
            raise ValueError(
                f"Problem '{self.name}': No variables added for optimization."
            )

        grps = self._vars.groupby("type")
        if "int" not in grps.groups.keys():
            return []
        else:
            return grps.get_group("int")["min"].to_numpy(FC.ITYPE)

    def max_values_int(self):
        """
        The maximal values of the integer variables.

        Use self.INT_INF for unbounded.

        Returns
        -------
        values: numpy.ndarray
            Maximal int values, shape: (n_vars_int,)

        """
        if self._vars is None:
            raise ValueError(
                f"Problem '{self.name}': No variables added for optimization."
            )

        grps = self._vars.groupby("type")
        if "int" not in grps.groups.keys():
            return []
        else:
            return grps.get_group("int")["max"].to_numpy(FC.ITYPE)

    def var_names_float(self):
        """
        The names of float variables.

        Returns
        -------
        names: list of str
            The names of the float variables

        """
        if self._vars is None:
            raise ValueError(
                f"Problem '{self.name}': No variables added for optimization."
            )

        grps = self._vars.groupby("type")
        if "float" not in grps.groups.keys():
            return []
        else:
            return grps.get_group("float")["name"].tolist()

    def initial_values_float(self):
        """
        The initial values of the float variables.

        Returns
        -------
        values: numpy.ndarray
            Initial float values, shape: (n_vars_float,)

        """
        if self._vars is None:
            raise ValueError(
                f"Problem '{self.name}': No variables added for optimization."
            )

        grps = self._vars.groupby("type")
        if "float" not in grps.groups.keys():
            return []
        else:
            return grps.get_group("float")["init"].to_numpy(FC.DTYPE)

    def min_values_float(self):
        """
        The minimal values of the float variables.

        Use -numpy.inf for unbounded.

        Returns
        -------
        values: numpy.ndarray
            Minimal float values, shape: (n_vars_float,)

        """
        if self._vars is None:
            raise ValueError(
                f"Problem '{self.name}': No variables added for optimization."
            )

        grps = self._vars.groupby("type")
        if "float" not in grps.groups.keys():
            return []
        else:
            return grps.get_group("float")["min"].to_numpy(FC.DTYPE)

    def max_values_float(self):
        """
        The maximal values of the float variables.

        Use numpy.inf for unbounded.

        Returns
        -------
        values: numpy.ndarray
            Maximal float values, shape: (n_vars_float,)

        """
        if self._vars is None:
            raise ValueError(
                f"Problem '{self.name}': No variables added for optimization."
            )

        grps = self._vars.groupby("type")
        if "float" not in grps.groups.keys():
            return []
        else:
            return grps.get_group("float")["max"].to_numpy(FC.DTYPE)

    def opt2farm_vars_individual(self, vars_int, vars_float):
        """
        Translates optimization variables to farm variables

        Parameters
        ----------
        vars_int: numpy.ndarray
            The integer optimization variable values,
            shape: (n_vars_int,)
        vars_float: numpy.ndarray
            The float optimization variable values,
            shape: (n_vars_float,)

        Returns
        -------
        farm_vars: dict
            The foxes farm variables. Key: var name,
            value: numpy.ndarray with values, shape:
            (n_states, n_sel_turbines)

        """
        n_states = self.algo.n_states
        n_sturb = self.n_sel_turbines

        farm_vars = {}
        grps = self._vars.groupby(["type", "var", "level"])
        for (typ, var, level), g in grps:
            src = vars_int if typ == "int" else vars_float
            i0 = g.index[0]
            i1 = g.index[-1]
            data = src[np.s_[i0 : i1 + 1]]

            if level == "uniform":
                farm_vars[var] = np.full((n_states, n_sturb), data[0], dtype=FC.DTYPE)

            elif level == "state":
                farm_vars[var] = np.full((n_states, n_sturb), np.nan, dtype=FC.DTYPE)
                if np.all(g["state"] == np.arange(n_states)):
                    farm_vars[var][:] = data[:, None]
                else:
                    farm_vars[var][g["state"]] = data[:, None]

            elif level == "turbine":
                farm_vars[var] = np.full((n_states, n_sturb), np.nan, dtype=FC.DTYPE)
                if np.all(g["sel_turbine"] == np.arange(n_sturb)):
                    farm_vars[var][:] = data[None, :]
                else:
                    farm_vars[var][:, g["sel_turbine"]] = data[None, :]

            elif level == "state-turbine":
                farm_vars[var] = np.full((n_states, n_sturb), np.nan, dtype=FC.DTYPE)
                farm_vars[var][g["state"], g["sel_turbine"]] = data

            else:
                raise ValueError(
                    f"Problem '{self.name}': Unknown level '{level}' encountered for variable '{var}'. Valid choices: uniform, state, turbine, state-turbine"
                )

        return farm_vars

    def opt2farm_vars_population(self, vars_int, vars_float, n_states):
        """
        Translates optimization variables to farm variables

        Parameters
        ----------
        vars_int: numpy.ndarray
            The integer optimization variable values,
            shape: (n_pop, n_vars_int)
        vars_float: numpy.ndarray
            The float optimization variable values,
            shape: (n_pop, n_vars_float)
        n_states: int
            The number of original (non-pop) states

        Returns
        -------
        farm_vars: dict
            The foxes farm variables. Key: var name,
            value: numpy.ndarray with values, shape:
            (n_pop, n_states, n_sel_turbines)

        """
        n_pop = vars_float.shape[0]
        n_sturb = self.n_sel_turbines

        farm_vars = {}
        grps = self._vars.groupby(["type", "var", "level"])
        for (typ, var, level), g in grps:
            src = vars_int if typ == "int" else vars_float
            i0 = g.index[0]
            i1 = g.index[-1]
            data = src[:, np.s_[i0 : i1 + 1]]

            if level == "uniform":
                farm_vars[var] = np.full(
                    (n_pop, n_states, n_sturb), np.nan, dtype=FC.DTYPE
                )
                farm_vars[var][:] = data[:, 0, None, None]

            elif level == "state":
                farm_vars[var] = np.full(
                    (n_pop, n_states, n_sturb), np.nan, dtype=FC.DTYPE
                )
                if np.all(g["state"] == np.arange(n_states)):
                    farm_vars[var][:] = data[:, :, None]
                else:
                    farm_vars[var][:, g["state"]] = data[:, :, None]

            elif level == "turbine":
                farm_vars[var] = np.full(
                    (n_pop, n_states, n_sturb), np.nan, dtype=FC.DTYPE
                )
                if np.all(g["sel_turbine"] == np.arange(n_sturb)):
                    farm_vars[var][:] = data[:, None, :]
                else:
                    farm_vars[var][:, :, g["sel_turbine"]] = data[:, None, :]

            elif level == "state-turbine":
                farm_vars[var] = np.full(
                    (n_pop, n_states, n_sturb), np.nan, dtype=FC.DTYPE
                )
                farm_vars[var][:, g["state"], g["sel_turbine"]] = data

            else:
                raise ValueError(
                    f"Problem '{self.name}': Unknown level '{level}' encountered for variable '{var}'. Valid choices: uniform, state, turbine, state-turbine"
                )

        return farm_vars



"""
Wake superposition models.
"""

from .ws_linear import WSLinear, WSLinearLocal
from .ws_quadratic import WSQuadratic, WSQuadraticLocal
from .ws_pow import WSPow, WSPowLocal
from .ws_max import WSMax, WSMaxLocal
from .ws_product import WSProduct
from .ti_linear import TILinear
from .ti_quadratic import TIQuadratic
from .ti_pow import TIPow
from .ti_max import TIMax



import numpy as np
from scipy.spatial.distance import cdist

from foxes.core import WakeFrame
from foxes.utils import wd2uv
from foxes.core.data import TData
import foxes.variables as FV
import foxes.constants as FC
from foxes.algorithms import Sequential


class SeqDynamicWakes(WakeFrame):
    """
    Dynamic wakes for the sequential algorithm.

    Attributes
    ----------
    cl_ipars: dict
        Interpolation parameters for centre line
        point interpolation
    dt_min: float, optional
        The delta t value in minutes,
        if not from timeseries data

    :group: models.wake_frames.sequential

    """

    def __init__(self, cl_ipars={}, dt_min=None):
        """
        Constructor.

        Parameters
        ----------
        cl_ipars: dict
            Interpolation parameters for centre line
            point interpolation
        dt_min: float, optional
            The delta t value in minutes,
            if not from timeseries data

        """
        super().__init__()
        self.cl_ipars = cl_ipars
        self.dt_min = dt_min

    def __repr__(self):
        return f"{type(self).__name__}(dt_min={self.dt_min})"

    def initialize(self, algo, verbosity=0):
        """
        Initializes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        """
        super().initialize(algo, verbosity)
        if not isinstance(algo, Sequential):
            raise TypeError(
                f"Incompatible algorithm type {type(algo).__name__}, expecting {Sequential.__name__}"
            )

        # determine time step:
        times = np.asarray(algo.states.index())
        if self.dt_min is None:
            if not np.issubdtype(times.dtype, np.datetime64):
                raise TypeError(
                    f"{self.name}: Expecting state index of type np.datetime64, found {times.dtype}"
                )
            elif len(times) == 1:
                raise KeyError(
                    f"{self.name}: Expecting 'dt_min' for single step timeseries"
                )
            self._dt = (
                (times[1:] - times[:-1]).astype("timedelta64[s]").astype(FC.ITYPE)
            )
        else:
            n = max(len(times) - 1, 1)
            self._dt = np.full(n, self.dt_min * 60, dtype="timedelta64[s]").astype(
                FC.ITYPE
            )

        # init wake traces data:
        self._traces_p = np.zeros((algo.n_states, algo.n_turbines, 3), dtype=FC.DTYPE)
        self._traces_v = np.zeros((algo.n_states, algo.n_turbines, 3), dtype=FC.DTYPE)
        self._traces_l = np.zeros((algo.n_states, algo.n_turbines), dtype=FC.DTYPE)

    def calc_order(self, algo, mdata, fdata):
        """ "
        Calculates the order of turbine evaluation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data

        Returns
        -------
        order: numpy.ndarray
            The turbine order, shape: (n_states, n_turbines)

        """
        # prepare:
        n_states = fdata.n_states
        n_turbines = algo.n_turbines
        tdata = TData.from_points(points=fdata[FV.TXYH])

        # calculate streamline x coordinates for turbines rotor centre points:
        # n_states, n_turbines_source, n_turbines_target
        coosx = np.zeros((n_states, n_turbines, n_turbines), dtype=FC.DTYPE)
        for ti in range(n_turbines):
            coosx[:, ti, :] = self.get_wake_coos(algo, mdata, fdata, tdata, ti)[
                :, :, 0, 0
            ]

        # derive turbine order:
        # TODO: Remove loop over states
        order = np.zeros((n_states, n_turbines), dtype=FC.ITYPE)
        for si in range(n_states):
            order[si] = np.lexsort(keys=coosx[si])

        return order

    def get_wake_coos(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
    ):
        """
        Calculate wake coordinates of rotor points.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwnd order

        Returns
        -------
        wake_coos: numpy.ndarray
            The wake frame coordinates of the evaluation
            points, shape: (n_states, n_targets, n_tpoints, 3)

        """
        # prepare:
        n_states = 1
        n_targets = tdata.n_targets
        n_tpoints = tdata.n_tpoints
        n_points = n_targets * n_tpoints
        points = tdata[FC.TARGETS].reshape(n_states, n_points, 3)
        counter = algo.states.counter
        N = counter + 1

        # new wake starts at turbine:
        self._traces_p[counter, downwind_index] = fdata[FV.TXYH][0, downwind_index]
        self._traces_l[counter, downwind_index] = 0

        # transport wakes that originate from previous time steps:
        if counter > 0:
            dxyz = self._traces_v[:counter, downwind_index] * self._dt[:counter, None]
            self._traces_p[:counter, downwind_index] += dxyz
            self._traces_l[:counter, downwind_index] += np.linalg.norm(dxyz, axis=-1)

        # compute wind vectors at wake traces:
        # TODO: dz from U_z is missing here
        hpdata = {
            v: np.zeros((1, N, 1), dtype=FC.DTYPE)
            for v in algo.states.output_point_vars(algo)
        }
        hpdims = {v: (FC.STATE, FC.TARGET, FC.TPOINT) for v in hpdata.keys()}
        hpdata = TData.from_points(
            points=self._traces_p[None, :N, downwind_index],
            data=hpdata,
            dims=hpdims,
        )
        res = algo.states.calculate(algo, mdata, fdata, hpdata)
        self._traces_v[:N, downwind_index, :2] = wd2uv(
            res[FV.WD][0, :, 0], res[FV.WS][0, :, 0]
        )
        del hpdata, hpdims, res

        # project:
        dists = cdist(points[0], self._traces_p[:N, downwind_index])
        tri = np.argmin(dists, axis=1)
        del dists
        wcoos = np.full((n_states, n_points, 3), 1e20, dtype=FC.DTYPE)
        wcoos[0, :, 2] = points[0, :, 2] - fdata[FV.TXYH][:, downwind_index][0, None, 2]
        delp = points[0, :, :2] - self._traces_p[tri, downwind_index, :2]
        nx = self._traces_v[tri, downwind_index, :2]
        nx /= np.linalg.norm(nx, axis=1)[:, None]
        ny = np.concatenate([-nx[:, 1, None], nx[:, 0, None]], axis=1)
        wcoos[0, :, 0] = (
            np.einsum("pd,pd->p", delp, nx) + self._traces_l[tri, downwind_index]
        )
        wcoos[0, :, 1] = np.einsum("pd,pd->p", delp, ny)

        # turbines that cause wake:
        tdata[FC.STATE_SOURCE_ORDERI] = downwind_index

        # states that cause wake for each target point:
        tdata.add(
            FC.STATES_SEL,
            tri[None, :].reshape(n_states, n_targets, n_tpoints),
            (FC.STATE, FC.TARGET, FC.TPOINT),
        )

        return wcoos.reshape(n_states, n_targets, n_tpoints, 3)

    def get_wake_modelling_data(
        self,
        algo,
        variable,
        downwind_index,
        fdata,
        tdata,
        target,
        states0=None,
        upcast=False,
    ):
        """
        Return data that is required for computing the
        wake from source turbines to evaluation points.

        Parameters
        ----------
        algo: foxes.core.Algorithm, optional
            The algorithm, needed for data from previous iteration
        variable: str
            The variable, serves as data key
        downwind_index: int, optional
            The index in the downwind order
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        target: str, optional
            The dimensions identifier for the output,
            FC.STATE_TARGET, FC.STATE_TARGET_TPOINT
        states0: numpy.ndarray, optional
            The states of wake creation
        upcast: bool
            Flag for ensuring targets dimension,
            otherwise dimension 1 is entered

        Returns
        -------
        data: numpy.ndarray
            Data for wake modelling, shape:
            (n_states, n_turbines) or (n_states, n_target)
        dims: tuple
            The data dimensions

        """
        if states0 is None and FC.STATE_SOURCE_ORDERI in tdata:
            # from previous iteration:
            if downwind_index != tdata[FC.STATE_SOURCE_ORDERI]:
                raise ValueError(
                    f"Model '{self.name}': Mismatch of '{FC.STATE_SOURCE_ORDERI}'. Expected {tdata[FC.STATE_SOURCE_ORDERI]}, got {downwind_index}"
                )

            n_states = 1
            n_targets = tdata.n_targets
            n_tpoints = tdata.n_tpoints
            n_points = n_targets * n_tpoints

            s = tdata[FC.STATES_SEL][0].reshape(n_points)
            data = algo.farm_results[variable].to_numpy()
            data = data[s, downwind_index].reshape(n_states, n_targets, n_tpoints)

            if target == FC.STATE_TARGET:
                if n_tpoints == 1:
                    data = data[:, :, 0]
                else:
                    data = np.einsum("stp,p->st", data, tdata[FC.TWEIGHTS])
                return data, (FC.STATE, FC.TARGET)
            elif target == FC.STATE_TARGET_TPOINT:
                return data, (FC.STATE, FC.TARGET, FC.TPOINT)
            else:
                raise ValueError(
                    f"Cannot handle target '{target}', choices are {FC.STATE_TARGET}, {FC.STATE_TARGET_TPOINT}"
                )

        else:
            return super().get_wake_modelling_data(
                algo, variable, downwind_index, fdata, tdata, target, states0, upcast
            )

    def get_centreline_points(self, algo, mdata, fdata, downwind_index, x):
        """
        Gets the points along the centreline for given
        values of x.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        downwind_index: int
            The index in the downwind order
        x: numpy.ndarray
            The wake frame x coordinates, shape: (n_states, n_points)

        Returns
        -------
        points: numpy.ndarray
            The centreline points, shape: (n_states, n_points, 3)

        """
        raise NotImplementedError



import numpy as np
from copy import deepcopy

from foxes.core import FarmDataModel, TData


class FarmWakesCalculation(FarmDataModel):
    """
    This model calculates wakes effects on farm data.

    Attributes
    ----------
    urelax: foxes.algorithms.iterative.models.URelax
        The under-relaxation model

    :group: algorithms.iterative.models

    """

    def __init__(self, urelax=None):
        """
        Constructor.

        Parameters
        ----------
        urelax: foxes.algorithms.iterative.models.URelax, optional
            The under-relaxation model

        """
        super().__init__()
        self.urelax = urelax

    def output_farm_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        ovars = algo.rotor_model.output_farm_vars(
            algo
        ) + algo.farm_controller.output_farm_vars(algo)
        return list(dict.fromkeys(ovars))

    def sub_models(self):
        """
        List of all sub-models

        Returns
        -------
        smdls: list of foxes.core.Model
            All sub models

        """
        return [] if self.urelax is None else [self.urelax]

    def calculate(self, algo, mdata, fdata):
        """ "
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.Data
            The model data
        fdata: foxes.core.Data
            The farm data

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape (n_states, n_turbines)

        """
        # collect ambient rotor results and weights:
        rotor = algo.rotor_model
        weights = rotor.from_data_or_store(rotor.RWEIGHTS, algo, mdata)
        amb_res = rotor.from_data_or_store(rotor.AMBRES, algo, mdata)

        # generate all wake evaluation points
        # (n_states, n_order, n_rpoints)
        pwake2tdata = {}
        for wname, wmodel in algo.wake_models.items():
            pwake = algo.partial_wakes[wname]
            if pwake.name not in pwake2tdata:
                tpoints, tweights = pwake.get_wake_points(algo, mdata, fdata)
                pwake2tdata[pwake.name] = TData.from_tpoints(tpoints, tweights)

        def _get_wdata(tdatap, wdeltas, s):
            """Helper function for wake data extraction"""
            tdata = tdatap.get_slice(s, keep=True)
            wdelta = {v: d[s] for v, d in wdeltas.items()}
            return tdata, wdelta

        wake_res = deepcopy(amb_res)
        n_turbines = mdata.n_turbines
        for wname, wmodel in algo.wake_models.items():
            pwake = algo.partial_wakes[wname]
            gmodel = algo.ground_models[wname]
            tdatap = pwake2tdata[pwake.name]
            wdeltas = pwake.new_wake_deltas(algo, mdata, fdata, tdatap, wmodel)

            for oi in range(n_turbines):

                if oi > 0:
                    tdata, wdelta = _get_wdata(tdatap, wdeltas, np.s_[:, :oi])
                    gmodel.contribute_to_farm_wakes(
                        algo, mdata, fdata, tdata, oi, wdelta, wmodel, pwake
                    )

                if oi < n_turbines - 1:
                    tdata, wdelta = _get_wdata(tdatap, wdeltas, np.s_[:, oi + 1 :])
                    gmodel.contribute_to_farm_wakes(
                        algo, mdata, fdata, tdata, oi, wdelta, wmodel, pwake
                    )

            for oi in range(n_turbines):
                wres = gmodel.finalize_farm_wakes(
                    algo,
                    mdata,
                    fdata,
                    tdatap,
                    amb_res,
                    weights,
                    wdeltas,
                    wmodel,
                    oi,
                    pwake,
                )
                for v, d in wres.items():
                    if v in wake_res:
                        wake_res[v][:, oi] += d

            del pwake, tdatap, wdeltas

        rotor.eval_rpoint_results(algo, mdata, fdata, wake_res, weights)
        res = algo.farm_controller.calculate(algo, mdata, fdata, pre_rotor=False)
        if self.urelax is not None:
            res = self.urelax.calculate(algo, mdata, fdata, res)
        fdata.update(res)

        return {v: fdata[v] for v in self.output_farm_vars(algo)}



from pathlib import Path
import warnings


def parse_Pct_file_name(file_name):
    """
    Parse basic turbine data from file name

    Expected format: `[name]-[...]MW-D[...]-H[...].csv`

    Parameters
    ----------
    file_name: str or pathlib.Path
        Path to the file

    Returns
    -------
    parsed_data: dict
        dict with data parsed from file name

    :group: data

    """
    sname = Path(file_name).stem
    pars = {"name": sname.split(".")[0]}

    i = sname.find(".")
    if i >= 0:
        if "-" in sname[i:]:
            warnings.warn(
                f"Illegal use of '.' in '{sname}', please replace by 'd' for float value dots. Parsing stopped."
            )
            return pars

    if "-" in sname and "_" in sname:
        warnings.warn(
            f"Illegal file name '{file_name}': Contains both '-' and '_'. Parsing stopped."
        )
        return pars

    if "-" in sname:
        pieces = sname.split("-")
    elif "_" in sname:
        pieces = sname.split("_")

    pars["name"] = pieces[0]
    pieces = pieces[1:]
    for p in pieces:
        if p[-1] == "W":
            if p[-2] == "k":
                pars["P_nominal"] = float(p[:-2].replace("d", "."))
            elif p[-2] == "M":
                pars["P_nominal"] = 1.0e3 * float(p[:-2].replace("d", "."))
            elif p[-2] == "G":
                pars["P_nominal"] = 1.0e6 * float(p[:-2].replace("d", "."))
            else:
                pars["P_nominal"] = 1.0e-3 * float(p[:-1].replace("d", "."))
            pars["name"] += "-" + p

        elif p[0] == "D":
            pars["D"] = float(p[1:].replace("d", "."))
        elif p[0] == "H":
            pars["H"] = float(p[1:].replace("d", "."))
        else:
            pars["name"] += "-" + p

    return pars


def parse_Pct_two_files(file_name_A, file_name_B):
    """
    Parse basic turbine data from file names

    Expected format: `[name]-[...]MW-D[...]-H[...].csv`

    Parameters
    ----------
    file_name_A: str or pathlib.Path
        Path to the first file
    file_name_B: str or pathlib.Path
        Path to the second file

    Returns
    -------
    parsed_data: dict
        dict with data parsed from file name

    :group: data

    """
    pars_A = parse_Pct_file_name(file_name_A)
    pars_B = parse_Pct_file_name(file_name_B)
    name = pars_A["name"]
    name_ct = pars_B["name"]
    i = 0
    while len(name) > i and len(name_ct) > i and name[i] == name_ct[i]:
        i += 1
    if i > 0 and name[i - 1] == "-":
        i -= 1
    if i < 1:
        raise ValueError(
            f"Turbine type name not deducible. From file A: '{name}', from file B: '{name_ct}'"
        )
    pars_A["name"] = name[:i]
    pars_B["name"] = name[:i]
    if pars_A != pars_B:
        raise ValueError(
            f"Data parsing from file names failed. File '{file_name_A}' gave '{pars_A}', file '{file_name_B}' gave '{pars_B}'"
        )

    return pars_A



import numpy as np
import pandas as pd
from scipy.interpolate import interpn

from foxes.core import TurbineType
from foxes.utils import PandasFileHelper
from foxes.data import PCTCURVE, parse_Pct_two_files
import foxes.variables as FV
import foxes.constants as FC


class WsTI2PCtFromTwo(TurbineType):
    """
    Calculate turbulent intensity dependent power
    and ct values, as given by two individual
    files.

    The structure of each file is:
    ws,0.05,0.06,0.07,...

    The first column represents wind speed in m/s
    and the subsequent columns are TI values
    (not neccessarily in order).

    Attributes
    ----------
    source_P: str or pandas.DataFrame
        The file path for the power curve, static name, or data
    source_ct: str or pandas.DataFrame
        The file path for the ct curve, static name, or data
    WSCT: str
        The wind speed variable for ct lookup
    WSP: str
        The wind speed variable for power lookup
    rpars_P: dict, optional
        Parameters for pandas power file reading
    rpars_ct: dict, optional
        Parameters for pandas ct file reading
    ipars_P: dict, optional
        Parameters for scipy.interpolate.interpn()
    ipars_ct: dict, optional
        Parameters for scipy.interpolate.interpn()
    rho: float
        The air densitiy for which the data is valid
        or None for no correction

    :group: models.turbine_types

    """

    def __init__(
        self,
        data_source_P,
        data_source_ct,
        rho=None,
        p_ct=1.0,
        p_P=1.88,
        var_ws_ct=FV.REWS2,
        var_ws_P=FV.REWS3,
        pd_file_read_pars_P={},
        pd_file_read_pars_ct={},
        interpn_pars_P=None,
        interpn_pars_ct=None,
        **parameters,
    ):
        """
        Constructor.

        Parameters
        ----------
        data_source_P: str or pandas.DataFrame
            The file path for the power curve, static name, or data
        data_source_ct: str or pandas.DataFrame
            The file path for the ct curve, static name, or data
        rho: float, optional
            The air densitiy for which the data is valid
            or None for no correction
        p_ct: float
            The exponent for yaw dependency of ct
        p_P: float
            The exponent for yaw dependency of P
        var_ws_ct: str
            The wind speed variable for ct lookup
        var_ws_P: str
            The wind speed variable for power lookup
        pd_file_read_pars_P:  dict
            Parameters for pandas power file reading
        pd_file_read_pars_ct:  dict
            Parameters for pandas ct file reading
        interpn_pars_P: dict, optional
            Parameters for scipy.interpolate.interpn()
        interpn_pars_ct: dict, optional
            Parameters for scipy.interpolate.interpn()
        parameters: dict, optional
            Additional parameters for TurbineType class

        """
        if not isinstance(data_source_P, pd.DataFrame) or not isinstance(
            data_source_ct, pd.DataFrame
        ):
            pars = parse_Pct_two_files(data_source_P, data_source_ct)
        else:
            pars = parameters
        super().__init__(**pars)

        self.source_P = data_source_P
        self.source_ct = data_source_ct
        self.rho = rho
        self.p_ct = p_ct
        self.p_P = p_P
        self.WSCT = var_ws_ct
        self.WSP = var_ws_P
        self.rpars_P = pd_file_read_pars_P
        self.rpars_ct = pd_file_read_pars_ct
        self.ipars_P = interpn_pars_P
        self.ipars_ct = interpn_pars_ct

        if self.ipars_P is None:
            self.ipars_P = dict(method="linear", bounds_error=True, fill_value=0.0)
        if self.ipars_ct is None:
            self.ipars_ct = dict(method="linear", bounds_error=True, fill_value=0.0)

        self._P = None
        self._ct = None

    def __repr__(self):
        a = f"D={self.D}, H={self.H}, P_nominal={self.P_nominal}, P_unit={self.P_unit}, rho={self.rho}"
        return f"{type(self).__name__}({a})"

    def output_farm_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return [FV.P, FV.CT]

    def load_data(self, algo, verbosity=0):
        """
        Load and/or create all model data that is subject to chunking.

        Such data should not be stored under self, for memory reasons. The
        data returned here will automatically be chunked and then provided
        as part of the mdata object during calculations.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        idata: dict
            The dict has exactly two entries: `data_vars`,
            a dict with entries `name_str -> (dim_tuple, data_ndarray)`;
            and `coords`, a dict with entries `dim_name_str -> dim_array`

        """
        # read power curve:
        if isinstance(self.source_P, pd.DataFrame):
            data = self.source_P
        else:
            fpath = algo.dbook.get_file_path(PCTCURVE, self.source_P, check_raw=True)
            pars = {"index_col": 0}
            pars.update(self.rpars_P)
            data = PandasFileHelper.read_file(fpath, **pars)

        data.sort_index(inplace=True)
        data.columns = data.columns.astype(FC.DTYPE)
        self._ws_P = data.index.to_numpy(FC.DTYPE)
        self._ti_P = np.sort(data.columns.to_numpy())
        self._P = data[self._ti_P].to_numpy(FC.DTYPE)

        # read ct curve:
        if isinstance(self.source_ct, pd.DataFrame):
            data = self.source_ct
        else:
            fpath = algo.dbook.get_file_path(PCTCURVE, self.source_ct, check_raw=True)
            pars = {"index_col": 0}
            pars.update(self.rpars_ct)
            data = PandasFileHelper.read_file(fpath, **pars)

        data.sort_index(inplace=True)
        data.columns = data.columns.astype(FC.DTYPE)
        self._ws_ct = data.index.to_numpy(FC.DTYPE)
        self._ti_ct = np.sort(data.columns.to_numpy())
        self._ct = data[self._ti_ct].to_numpy(FC.DTYPE)

        return super().load_data(algo, verbosity)

    def _bounds_info(self, target, qts):
        """Helper function for printing bounds info"""

        print(f"\nBOUNDS INFO FOR TARGET {target}")
        WS = self.WSP if target == FV.P else self.WSCT
        ws = self._ws_P if target == FV.P else self._ws_ct
        ti = self._ti_P if target == FV.P else self._ti_ct
        print(f"  {WS}: min = {np.min(ws):.4f}, max = {np.max(ws):.4f}")
        print(f"  {FV.TI}: min = {np.min(ti):.4f}, max = {np.max(ti):.4f}")

        print(f"DATA INFO FOR TARGET {target}")
        ws = qts[:, 0]
        ti = qts[:, 1]
        print(f"  {WS}: min = {np.min(ws):.4f}, max = {np.max(ws):.4f}")
        print(f"  {FV.TI}: min = {np.min(ti):.4f}, max = {np.max(ti):.4f}")
        print()

    def calculate(self, algo, mdata, fdata, st_sel):
        """ "
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        st_sel: numpy.ndarray of bool
            The state-turbine selection,
            shape: (n_states, n_turbines)

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape (n_states, n_turbines)

        """

        # calculate P:
        st_sel_P = (
            st_sel
            & (fdata[self.WSP] >= self._ws_P[0])
            & (fdata[self.WSP] <= self._ws_P[-1])
        )
        st_sel_P0 = st_sel & ~st_sel_P
        if np.any(st_sel_P0):
            fdata[FV.P][st_sel_P0] = 0
        if np.any(st_sel_P):
            # prepare interpolation:
            n_sel = np.sum(st_sel_P)
            qts = np.zeros((n_sel, 2), dtype=FC.DTYPE)  # ws, ti
            qts[:, 0] = fdata[self.WSP][st_sel_P]
            qts[:, 1] = fdata[FV.TI][st_sel_P]

            # apply air density correction:
            if self.rho is not None:
                # correct wind speed by air density, such
                # that in the partial load region the
                # correct value is reconstructed:
                rho = fdata[FV.RHO][st_sel]
                qts[:, 0] *= (self.rho / rho) ** (1.0 / 3.0)
                del rho

            # apply yaw corrections:
            if FV.YAWM in fdata and self.p_P is not None:
                # calculate corrected wind speed wsc,
                # gives ws**3 * cos**p_P in partial load region
                # and smoothly deals with full load region:
                yawm = fdata[FV.YAWM][st_sel_P]
                if np.any(np.isnan(yawm)):
                    raise ValueError(
                        f"{self.name}: Found NaN values for variable '{FV.YAWM}'. Maybe change order in turbine_models?"
                    )
                cosm = np.cos(yawm / 180 * np.pi)
                qts[:, 0] *= (cosm**self.p_P) ** (1.0 / 3.0)
                del yawm, cosm

            # run interpolation:
            try:
                fdata[FV.P][st_sel_P] = interpn(
                    (self._ws_P, self._ti_P), self._P, qts, **self.ipars_P
                )
            except ValueError as e:
                self._bounds_info(FV.P, qts)
                raise e
        del st_sel_P, st_sel_P0

        # calculate ct:
        st_sel_ct = (
            st_sel
            & (fdata[self.WSCT] >= self._ws_P[0])
            & (fdata[self.WSCT] <= self._ws_P[-1])
        )
        st_sel_ct0 = st_sel & ~st_sel_ct
        if np.any(st_sel_ct0):
            fdata[FV.CT][st_sel_ct0] = 0
        if np.any(st_sel_ct):
            # prepare interpolation:
            n_sel = np.sum(st_sel_ct)
            qts = np.zeros((n_sel, 2), dtype=FC.DTYPE)  # ws, ti
            qts[:, 0] = fdata[self.WSP][st_sel_ct]
            qts[:, 1] = fdata[FV.TI][st_sel_ct]

            # apply air density correction:
            if self.rho is not None:
                # correct wind speed by air density, such
                # that in the partial load region the
                # correct value is reconstructed:
                rho = fdata[FV.RHO][st_sel]
                qts[:, 0] *= (self.rho / rho) ** 0.5
                del rho

            # apply yaw corrections:
            if FV.YAWM in fdata and self.p_ct is not None:
                # calculate corrected wind speed wsc,
                # gives ws**3 * cos**p_P in partial load region
                # and smoothly deals with full load region:
                yawm = fdata[FV.YAWM][st_sel_ct]
                if np.any(np.isnan(yawm)):
                    raise ValueError(
                        f"{self.name}: Found NaN values for variable '{FV.YAWM}'. Maybe change order in turbine_models?"
                    )
                cosm = np.cos(yawm / 180 * np.pi)
                qts[:, 0] *= (cosm**self.p_ct) ** 0.5
                del yawm, cosm

            # run interpolation:
            try:
                fdata[FV.CT][st_sel_ct] = interpn(
                    (self._ws_ct, self._ti_ct), self._ct, qts, **self.ipars_ct
                )
            except ValueError as e:
                self._bounds_info(FV.CT, qts)
                raise e

        return {v: fdata[v] for v in self.output_farm_vars(algo)}

    def finalize(self, algo, verbosity=0):
        """
        Finalizes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level

        """
        del self._ws_P, self._ti_P, self._ws_ct, self._ti_ct
        self._P = None
        self._ct = None
        super().finalize(algo, verbosity)



def all_subclasses(cls):
    """
    Searches all classes derived from some
    base class.

    Parameters
    ----------
    cls: class
        The base class

    Returns
    -------
    list of class:
        The derived classes

    :group: utils

    """
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in all_subclasses(c)]
    )



import pandas as pd

from foxes.core import Turbine


def add_from_csv(
    farm,
    data_source,
    col_index=None,
    col_name=None,
    col_x="x",
    col_y="y",
    col_H=None,
    col_D=None,
    col_id=None,
    cols_models_pre=None,
    cols_models_post=None,
    turbine_base_name="T",
    turbine_ids=None,
    turbine_base_name_count_shift=False,
    verbosity=1,
    **turbine_parameters,
):
    """
    Add turbines to wind farm via csv input file.

    Parameters
    ----------
    farm: foxes.WindFarm
        The wind farm
    data_source: str or pandas.DataFrame
        The input csv file or data source
    col_index: str, optional
        The index column, or None
    col_name: str, optional
        The name column, or None
    col_x: str, optional
        The x column
    col_y: str, optional
        The y column
    col_H: str, optional
        The hub height column
    col_D: str, optional
        The rotor diameter column
    col_id: str, optional
        The id column
    cols_models_pre: list of str, optional
        The turbine model columns, entered before
        turbine_models
    cols_models_post: list of str, optional
        The turbine model columns, entered after
        turbine_models
    turbine_base_name: str, optional
        The turbine base name, only used
        if col_name is None
    turbine_ids: list, optional
        The turbine ids, or None for
        index
    turbine_base_name_count_shift: bool, optional
        Start turbine names by 1 instead of 0
    verbosity: int
        The verbosity level, 0 = silent
    turbine_parameters: dict, optional
        Additional parameters are forwarded to the WindFarm.add_turbine().

    :group: input.farm_layout

    """

    if isinstance(data_source, pd.DataFrame):
        data = data_source
    else:
        if verbosity:
            print("Reading file", data_source)
        data = pd.read_csv(data_source, index_col=col_index)

    tmodels = turbine_parameters.pop("turbine_models", [])
    H = turbine_parameters.pop("H", None)
    D = turbine_parameters.pop("D", None)

    for i in data.index:
        s = 1 if turbine_base_name_count_shift else 0
        tname = (
            f"{turbine_base_name}{i+s}" if col_name is None else data.loc[i, col_name]
        )
        txy = data.loc[i, [col_x, col_y]].values

        if turbine_ids is not None:
            tid = turbine_ids[i]
        elif col_id is not None:
            tid = data.loc[i, col_id]
        else:
            tid = None

        hmodels = (
            [] if cols_models_pre is None else data.loc[i, cols_models_pre].tolist()
        )
        hmodels += tmodels
        hmodels += (
            [] if cols_models_post is None else data.loc[i, cols_models_post].tolist()
        )

        farm.add_turbine(
            Turbine(
                name=tname,
                index=tid,
                xy=txy,
                H=H if col_H not in data.columns else data.loc[i, col_H],
                D=D if col_D not in data.columns else data.loc[i, col_D],
                turbine_models=hmodels,
                **turbine_parameters,
            ),
            verbosity=verbosity,
        )



"""
Output tools and functions.
"""

from .round import round_defaults
from .output import Output
from .farm_layout import FarmLayoutOutput
from .farm_results_eval import FarmResultsEval
from .rose_plot import RosePlotOutput, StatesRosePlotOutput
from .results_writer import ResultsWriter
from .state_turbine_map import StateTurbineMap
from .turbine_type_curves import TurbineTypeCurves
from .animation import Animator
from .calc_points import PointCalculator
from .slice_data import SliceData
from .rotor_point_plots import RotorPointPlot
from .state_turbine_table import StateTurbineTable

from .flow_plots_2d import FlowPlots2D, SeqFlowAnimationPlugin
from . import grids



import numpy as np
from abc import abstractmethod

from foxes.models.wake_models.axisymmetric import AxisymmetricWakeModel


class GaussianWakeModel(AxisymmetricWakeModel):
    """
    Abstract base class for Gaussian wake models.

    :group: models.wake_models

    """

    @abstractmethod
    def calc_amplitude_sigma(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        x,
    ):
        """
        Calculate the amplitude and the sigma,
        both depend only on x (not on r).

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index in the downwind order
        x: numpy.ndarray
            The x values, shape: (n_states, n_targets)

        Returns
        -------
        amsi: tuple
            The amplitude and sigma, both numpy.ndarray
            with shape (n_st_sel,)
        st_sel: numpy.ndarray of bool
            The state-target selection, for which the wake
            is non-zero, shape: (n_states, n_targets)

        """
        pass

    def calc_wakes_x_r(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        x,
        r,
    ):
        """
        Calculate wake deltas.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index in the downwind order
        x: numpy.ndarray
            The x values, shape: (n_states, n_targets)
        r: numpy.ndarray
            The radial values for each x value, shape:
            (n_states, n_targets, n_yz_per_target)

        Returns
        -------
        wdeltas: dict
            The wake deltas. Key: variable name str,
            value: numpy.ndarray, shape: (n_st_sel, n_r_per_x)
        st_sel: numpy.ndarray of bool
            The state-target selection, for which the wake
            is non-zero, shape: (n_states, n_targets)

        """
        amsi, st_sel = self.calc_amplitude_sigma(
            algo, mdata, fdata, tdata, downwind_index, x
        )
        wdeltas = {}
        rsel = r[st_sel]
        for v in amsi.keys():
            ampld, sigma = amsi[v]
            wdeltas[v] = ampld[:, None] * np.exp(-0.5 * (rsel / sigma[:, None]) ** 2)

        return wdeltas, st_sel



X = "X"
""" The x coordinate in m
:group: foxes.variables
"""

Y = "Y"
""" The y coordinate in 
:group: foxes.variables
"""

H = "H"
""" The height over ground in m
:group: foxes.variables
"""

D = "D"
""" The rotor diameter in m
:group: foxes.variables
"""

TXYH = "txyh"
""" The turbine rotor centre coordinate 
vector (x, y, height)
:group: foxes.variables
"""

WEIGHT = "weight"
""" The statistical weight of a state
:group: foxes.variables
"""

ORDER = "order"
""" The turbine order
:group: foxes.variables
"""

ORDER_INV = "order_inv"
""" The inverse of the turbine order
:group: foxes.variables
"""

ORDER_SSEL = "order_ssel"
""" The states selection for applying the order
:group: foxes.variables
"""

WS = "WS"
""" The wind speed in m/s
:group: foxes.variables
"""

WD = "WD"
""" The wind direction in degrees
:group: foxes.variables
"""

TI = "TI"
""" The turbulence intensity
:group: foxes.variables
"""

RHO = "RHO"
""" The air density in kg/m3
:group: foxes.variables
"""

YAW = "YAW"
""" The absolute yaw angle of a turbine in degrees
:group: foxes.variables
"""

YAWM = "YAWM"
""" The relative yaw angle of a turbine in degrees
:group: foxes.variables
"""

P = "P"
""" The power, unit depends on user choice
:group: foxes.variables
"""

MAX_P = "MAXP"
""" The maximal power, for derating/boost
:group: foxes.variables
"""

CT = "CT"
""" The thrust coefficient
:group: foxes.variables
"""

T = "T"
""" The temperature in Kelvin
:group: foxes.variables
"""

YLD = "YLD"
""" Yield in GWh/a
:group: foxes.variables
"""

EFF = "EFF"
""" Efficiency, equals P/AMB_P
:group: foxes.variables
"""

CAP = "CAP"
""" Capacity, equals P/P_nominal
:group: foxes.variables
"""


REWS = "REWS"
""" Rotor effective wind speed in m/s
:group: foxes.variables
"""

REWS2 = "REWS2"
""" Rotor effective wind speed in m/s,
calculated from second moment
:group: foxes.variables
"""

REWS3 = "REWS3"
""" Rotor effective wind speed in m/s,
calculated from third moment
:group: foxes.variables
"""


AMB_WS = "AMB_WS"
""" The ambient wind speed in m/s
:group: foxes.variables
"""

AMB_WD = "AMB_WD"
""" The ambient wind direction in degrees
:group: foxes.variables
"""

AMB_TI = "AMB_TI"
""" The ambient turbulence intensity
:group: foxes.variables
"""

AMB_RHO = "AMB_RHO"
""" The ambient air density in kg/m3
:group: foxes.variables
"""

AMB_YAW = "AMB_YAW"
""" The ambient absolute yaw angle of 
a turbine in degrees
:group: foxes.variables
"""

AMB_YAWM = "AMB_YAWM"
""" The ambient relative yaw angle of 
a turbine in degrees
:group: foxes.variables
"""

AMB_P = "AMB_P"
""" The ambient power, unit depends on user choice
:group: foxes.variables
"""

AMB_CT = "AMB_CT"
""" The ambient thrust coefficient
:group: foxes.variables
"""

AMB_T = "AMB_T"
""" The ambient temperature in Kelvin
:group: foxes.variables
"""

AMB_YLD = "AMB_YLD"
""" Ambient yield in GWh/a
:group: foxes.variables
"""

AMB_CAP = "AMB_CAP"
""" Ambient capacity, equals AMB_P/P_nominal
:group: foxes.variables
"""


AMB_REWS = "AMB_REWS"
""" Ambient rotor effective wind speed in m/s
:group: foxes.variables
"""

AMB_REWS2 = "AMB_REWS2"
""" Ambient rotor effective wind speed in m/s,
calculated from second moment
:group: foxes.variables
"""

AMB_REWS3 = "AMB_REWS3"
""" Ambient rotor effective wind speed in m/s,
calculated from third moment
:group: foxes.variables
"""


var2amb = {
    v: f"AMB_{v}"
    for v in [
        WS,
        WD,
        TI,
        RHO,
        YAW,
        YAWM,
        P,
        CT,
        T,
        REWS,
        REWS2,
        REWS3,
        YLD,
        CAP,
    ]
}
""" Mapping from variable to the corresponding
ambient variable
:group: foxes.variables
"""

amb2var = {a: v for v, a in var2amb.items()}
""" Mapping from ambient variable to the corresponding
waked variable
:group: foxes.variables
"""


K = "k"
""" Wake growth parameter
:group: foxes.variables
"""

KB = "kb"
""" KTI value for zero TI, K = KB + KTI*TI
:group: foxes.variables
"""

KTI = "kTI"
""" Factor between K and TI, K = KB + KTI*TI
:group: foxes.variables
"""


Z0 = "z0"
""" The roughness length in m
:group: foxes.variables
"""

MOL = "MOL"
""" The Monin-Ubukhof length in m
:group: foxes.variables
"""

SHEAR = "shear"
""" The shear exponent
:group: foxes.variables
"""


PA_ALPHA = "PA_alpha"
""" The alpha parameter of the PorteAgel wake model
:group: foxes.variables
"""

PA_BETA = "PA_beta"
""" The beta parameter of the PorteAgel wake model
:group: foxes.variables
"""



import numpy as np
from xarray import Dataset
from io import StringIO


def read_tab_file(fname, normalize=True):
    """
    Reads a tab file into a Dataset

    Parameters
    ----------
    fname: str
        The path to the tab file
    normalize: bool
        Normalize the frequencies such that
        they add to 1000 in each sector

    Returns
    -------
    out: xarray.Dataset
        The data

    :group: utils

    """

    header = []
    with open(fname, "r") as f:
        header.append(f.readline().replace("\t", " ").strip())
        header.append(f.readline().replace("\t", " ").strip())
        header.append(f.readline().replace("\t", " ").strip())
        sfreqs = f.readline().replace("\t", " ").strip()
        s = f.read().replace("\t", " ").strip()
    data = np.genfromtxt(StringIO(s))
    sfreqs = np.fromstring(sfreqs, sep=" ")

    descr = header[0]
    lat, lon, height = np.fromstring(header[1], sep=" ")
    nsec, a, b = np.fromstring(header[2], sep=" ")

    delta_wd = 360 / nsec
    out = Dataset(
        coords={"ws": data[:, 0], "wd": np.arange(0, 360.0, delta_wd)},
        data_vars={
            "wd_freq": (("wd",), sfreqs),
            "ws_freq": (("ws", "wd"), data[:, 1:]),
        },
        attrs={
            "description": descr,
            "latitude": lat,
            "longitude": lon,
            "height": height,
            "factor_ws": a,
            "shift_wd": b,
        },
    )

    if normalize:
        out["ws_freq"] *= 1000 / np.sum(data[:, 1:], axis=0)[None, :]
        out["wd_freq"] *= 100 / np.sum(sfreqs)

    return out


if __name__ == "__main__":

    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("fname", help="Path to the tab file")
    args = parser.parse_args()

    read_tab_file(args.fname)



import numpy as np

from foxes.core import WakeK
from foxes.models.wake_models.top_hat import TopHatWakeModel
import foxes.variables as FV
import foxes.constants as FC


class IECTIWake(TopHatWakeModel):
    """
    The TI wake model from IEC-64100-1-2005-8 (2005):

    Notes
    -----
    Reference:
    http://orbit.dtu.dk/files/3750291/2009_31.pdf
    v2: VolLuk: corrected implementation following: IEC-64100-1-2005-8
    (Appearently an error in the document by DTU)

    and the Frandsen wake TI model, from IEC-64100 (2019):
    http://orbit.dtu.dk/files/3750291/2009_31.pdf

    Attributes
    ----------
    wake_k: foxes.core.WakeK
        Handler for the wake growth parameter k

    :group: models.wake_models.ti

    """

    def __init__(
        self,
        superposition,
        opening_angle=21.6,
        iec_type="2019",
        induction="Betz",
        **wake_k,
    ):
        """
        Constructor.

        Parameters
        ----------
        superposition: str
            The TI wake superposition.
        opening_angle: float, optional
            The wake opening angle. The wake growth parameter k is calculated
            based on the wake opening angle.
        iec_type: str
            Either '2005' or '2019'/'Frandsen'
        wake_k: dict, optional
            Parameters for the WakeK class

        """
        super().__init__(superpositions={FV.TI: superposition}, induction=induction)

        if opening_angle is not None:
            if "k" in wake_k or "ka" in wake_k or "kb" in wake_k:
                raise KeyError(
                    f"Can handle 'opening_angle' or ('k', 'ka', 'kb') parameters, not both"
                )
            wake_k["k"] = float(np.tan(np.deg2rad(opening_angle / 2.0)))

        self.iec_type = iec_type
        self.wake_k = WakeK(**wake_k)

    def __repr__(self):
        iname = (
            self.induction if isinstance(self.induction, str) else self.induction.name
        )
        s = f"{type(self).__name__}"
        s += f"({self.superpositions[FV.TI]}, induction={iname}, "
        s += self.wake_k.repr() + ")"
        return s

    def sub_models(self):
        """
        List of all sub-models

        Returns
        -------
        smdls: list of foxes.core.Model
            All sub models

        """
        return [self.wake_k]

    def new_wake_deltas(self, algo, mdata, fdata, tdata):
        """
        Creates new empty wake delta arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data

        Returns
        -------
        wake_deltas: dict
            Key: variable name, value: The zero filled
            wake deltas, shape: (n_states, n_turbines, n_rpoints, ...)

        """
        return {FV.TI: np.zeros_like(tdata[FC.TARGETS][..., 0])}

    def calc_wake_radius(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        x,
        ct,
    ):
        """
        Calculate the wake radius, depending on x only (not r).

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index in the downwind order
        x: numpy.ndarray
            The x values, shape: (n_states, n_targets)
        ct: numpy.ndarray
            The ct values of the wake-causing turbines,
            shape: (n_states, n_targets)

        Returns
        -------
        wake_r: numpy.ndarray
            The wake radii, shape: (n_states, n_targets)

        """
        k = self.wake_k(
            FC.STATE_TARGET,
            algo=algo,
            fdata=fdata,
            tdata=tdata,
            upcast=False,
            downwind_index=downwind_index,
        )
        return k * x

    def calc_centreline(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        st_sel,
        x,
        wake_r,
        ct,
    ):
        """
        Calculate centre line results of wake deltas.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index in the downwind order
        st_sel: numpy.ndarray of bool
            The state-target selection, for which the wake
            is non-zero, shape: (n_states, n_targets)
        x: numpy.ndarray
            The x values, shape: (n_st_sel,)
        wake_r: numpy.ndarray
            The wake radii, shape: (n_st_sel,)
        ct: numpy.ndarray
            The ct values of the wake-causing turbines,
            shape: (n_st_sel,)

        Returns
        -------
        cl_del: dict
            The centre line wake deltas. Key: variable name str,
            varlue: numpy.ndarray, shape: (n_st_sel,)

        """
        # read D from extra data:
        D = self.get_data(
            FV.D,
            FC.STATE_TARGET,
            lookup="w",
            algo=algo,
            fdata=fdata,
            tdata=tdata,
            downwind_index=downwind_index,
            upcast=True,
        )[st_sel]

        # get ws:
        ws = self.get_data(
            FV.REWS,
            FC.STATE_TARGET,
            lookup="w",
            algo=algo,
            fdata=fdata,
            tdata=tdata,
            downwind_index=downwind_index,
            upcast=True,
        )[st_sel]

        # calculate wind deficit:
        if self.iec_type == "2005":
            cl_deltas = np.sqrt(0.9) / (1.5 + 0.3 * x / D * np.sqrt(ws))
        elif self.iec_type == "2019" or self.iec_type == "Frandsen":
            cl_deltas = 1.0 / (1.5 + 0.8 * x / D / np.sqrt(ct))
        else:
            raise TypeError(
                f"Type of IEC {self.iec_type} not found. Select '2015' or '2019'/'Frandsen'."
            )

        return {FV.TI: cl_deltas}



import numpy as np

from foxes.core import States, VerticalProfile
import foxes.variables as FV
import foxes.constants as FC


class SingleStateStates(States):
    """
    A single uniform state.

    Attributes
    ----------
    ws: float
        The wind speed
    wd: float
        The wind direction
    ti: float
        The TI value
    rho: float
        The air density
    profdicts: dict
        Key: output variable name str, Value: str or dict
        or `foxes.core.VerticalProfile`
    profdata: dict,
        Additional data for profiles

    :group: input.states

    """

    def __init__(self, ws=None, wd=None, ti=None, rho=None, profiles={}, **profdata):
        """
        Constructor.

        Parameters
        ----------
        ws: float, optional
            The wind speed
        wd: float, optional
            The wind direction
        ti: float, optional
            The TI value
        rho: float, optional
            The air density
        profiles: dict, optional
            Key: output variable name str, Value: str or dict
            or `foxes.core.VerticalProfile`
        profdata: dict, optional
            Additional data for profiles

        """
        super().__init__()
        self.ws = ws
        self.wd = wd
        self.ti = ti
        self.rho = rho
        self.profdicts = profiles
        self.profdata = profdata

        if (
            ws is None
            and wd is None
            and ti is None
            and rho is None
            and not len(profiles)
        ):
            raise KeyError(
                f"Expecting at least one parameter: ws, wd, ti, rho, profiles"
            )

    def sub_models(self):
        """
        List of all sub-models

        Returns
        -------
        smdls: list of foxes.core.Model
            Names of all sub models

        """
        return list(self._profiles.values())

    def initialize(self, algo, verbosity=0):
        """
        Initializes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        """
        self._profiles = {}
        for v, d in self.profdicts.items():
            if isinstance(d, str):
                self._profiles[v] = VerticalProfile.new(d)
            elif isinstance(d, VerticalProfile):
                self._profiles[v] = d
            elif isinstance(d, dict):
                t = d.pop("type")
                self._profiles[v] = VerticalProfile.new(t, **d)
            else:
                raise TypeError(
                    f"States '{self.name}': Wrong profile type '{type(d).__name__}' for variable '{v}'. Expecting VerticalProfile, str or dict"
                )
        super().initialize(algo, verbosity)

    def size(self):
        """
        The total number of states.

        Returns
        -------
        int:
            The total number of states

        """
        return 1

    def output_point_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        out = set()
        if self.ws is not None:
            out.add(FV.WS)
        if self.wd is not None:
            out.add(FV.WD)
        if self.ti is not None:
            out.add(FV.TI)
        if self.rho is not None:
            out.add(FV.RHO)
        out.update(list(self._profiles.keys()))

        return list(out)

    def weights(self, algo):
        """
        The statistical weights of all states.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        weights: numpy.ndarray
            The weights, shape: (n_states, n_turbines)

        """
        return np.ones((1, algo.n_turbines), dtype=FC.DTYPE)

    def calculate(self, algo, mdata, fdata, tdata):
        """ "
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape
            (n_states, n_targets, n_tpoints)

        """
        if self.ws is not None:
            tdata[FV.WS] = np.full(
                (tdata.n_states, tdata.n_targets, tdata.n_tpoints),
                self.ws,
                dtype=FC.DTYPE,
            )
        if self.wd is not None:
            tdata[FV.WD] = np.full(
                (tdata.n_states, tdata.n_targets, tdata.n_tpoints),
                self.wd,
                dtype=FC.DTYPE,
            )
        if self.ti is not None:
            tdata[FV.TI] = np.full(
                (tdata.n_states, tdata.n_targets, tdata.n_tpoints),
                self.ti,
                dtype=FC.DTYPE,
            )
        if self.rho is not None:
            tdata[FV.RHO] = np.full(
                (tdata.n_states, tdata.n_targets, tdata.n_tpoints),
                self.rho,
                dtype=FC.DTYPE,
            )

        if len(self._profiles):
            z = tdata[FC.TARGETS][..., 2]
            for k, v in self.profdata.items():
                tdata[k] = v
            for v, p in self._profiles.items():
                pres = p.calculate(tdata, z)
                tdata[v] = pres

        return {v: tdata[v] for v in self.output_point_vars(algo)}



import numpy as np

from foxes.output import Output
from foxes.utils import write_nc
import foxes.constants as FC
import foxes.variables as FV

from . import grids


class SliceData(Output):
    """
    Create data for horizontal or vertical 2D slices

    Attributes
    ----------
    algo: foxes.Algorithm
        The algorithm for point calculation
    farm_results: xarray.Dataset
        The farm results
    runner: foxes.utils.runners.Runner, optional
        The runner
    verbosity_delta: int
        Verbosity threshold for printing calculation info
            
    :group: output

    """

    def __init__(
        self, 
        algo, 
        farm_results, 
        runner=None, 
        verbosity_delta=1,
        **kwargs,
        ):
        """
        Constructor.

        Parameters
        ----------
        algo: foxes.Algorithm
            The algorithm for point calculation
        farm_results: xarray.Dataset
            The farm results
        runner: foxes.utils.runners.Runner, optional
            The runner
        verbosity_delta: int
            Verbosity threshold for printing calculation info
        kwargs: dict, optional
            Additional parameters for the base class

        """
        super().__init__(**kwargs)
        self.algo = algo
        self.fres = farm_results
        self.runner = runner
        self.verbosity_delta = verbosity_delta

    def _data_mod(
        self,
        a_pos,
        b_pos,
        c_pos,
        data,
        normalize_a,
        normalize_b,
        normalize_c,
        normalize_v,
        vmin,
        vmax,
    ):
        """Helper function for data modification"""
        if normalize_a is not None:
            a_pos /= normalize_a
        if normalize_b is not None:
            b_pos /= normalize_b
        if normalize_c is not None:
            c_pos /= normalize_c

        for v in data:
            if v in normalize_v:
                data[v] /= normalize_v[v]
            if v in vmin:
                data[v] = np.maximum(data[v], vmin[v])
            if v in vmax:
                data[v] = np.minimum(data[v], vmax[v])

        return a_pos, b_pos, c_pos, data

    def _write(self, format, data, fname, verbosity, **write_pars):
        """Helper function for file writing"""
        if fname is not None:
            if format == "numpy":
                fpath = self.get_fpath(fname)
                if verbosity > 0:
                    print("Writing file", fpath)
                wpars = dict(format="%.6f")
                wpars.update(write_pars)
                data.tofile(fpath, **wpars)

            elif format == "pandas":
                if verbosity > 0:
                    print("Writing file", fpath)
                self.write(fname, data, **write_pars)

            elif format == "xarray":
                write_nc(data, self.get_fpath(fname), verbosity=verbosity, **write_pars)

            else:
                raise ValueError(
                    f"Unknown data format '{format}', choices: numpy, pandas, xarray"
                )

    def _calc_mean_data(
        self,
        ori,
        data_format,
        variables,
        a_pos,
        b_pos,
        c_pos,
        g_pts,
        normalize_a,
        normalize_b,
        normalize_c,
        normalize_v,
        label_map,
        vmin,
        vmax,
        states_sel,
        states_isel,
        weight_turbine,
        to_file,
        write_pars,
        ret_states,
        verbosity,
        **kwargs,
    ):
        """Helper function for mean data calculation"""
        # calculate point results:
        point_results = grids.calc_point_results(
            algo=self.algo,
            farm_results=self.fres,
            g_pts=g_pts,
            sel={FC.STATE: states_sel} if states_sel is not None else None,
            isel={FC.STATE: states_isel} if states_isel is not None else None,
            verbosity=verbosity-self.verbosity_delta,
            **kwargs,
        )
        states = point_results[FC.STATE].to_numpy()
        if variables is None:
            variables = list(point_results.data_vars.keys())
        else:
            point_results.drop_vars(variables)
        del g_pts

        # take mean over states:
        weights = self.fres[FV.WEIGHT][:, weight_turbine].to_numpy()
        data = {
            v: np.einsum("s,sp->p", weights, point_results[v].to_numpy())
            for v in variables
        }
        del point_results

        # apply data modification:
        a_pos, b_pos, c_pos, data = self._data_mod(
            a_pos,
            b_pos,
            c_pos,
            data,
            normalize_a,
            normalize_b,
            normalize_c,
            normalize_v,
            vmin,
            vmax,
        )

        # translate to selected format:
        if data_format == "numpy":
            data = grids.np2np_p(data, a_pos, b_pos)
            self._write(data_format, data, to_file, verbosity, **write_pars)
        elif data_format == "pandas":
            data = grids.np2pd_p(data, a_pos, b_pos, ori, label_map)
            self._write(data_format, data, to_file, verbosity, **write_pars)
        elif data_format == "xarray":
            data = grids.np2xr_p(data, a_pos, b_pos, c_pos, ori, label_map)
            self._write(data_format, data, to_file, verbosity, **write_pars)
        else:
            raise ValueError(
                f"Unknown data format '{data_format}', choices: numpy, pandas, xarray"
            )

        return (data, states) if ret_states else data

    def get_mean_data_xy(
        self,
        resolution,
        variables=None,
        data_format="xarray",
        xmin=None,
        ymin=None,
        xmax=None,
        ymax=None,
        z=None,
        xspace=500.0,
        yspace=500.0,
        normalize_x=None,
        normalize_y=None,
        normalize_z=None,
        normalize_v={},
        label_map={},
        vmin={},
        vmax={},
        states_sel=None,
        states_isel=None,
        weight_turbine=0,
        to_file=None,
        write_pars={},
        ret_states=False,
        ret_grid=False,
        verbosity=0,
        **kwargs,
    ):
        """
        Creates mean data of 2D farm flow slices in a horizontal xy-plane.

        Parameters
        ----------
        resolution: float
            The resolution in m
        variables: list of str, optional
            The variables, or None for all
        data_format: str
            The output data format: numpy, pandas, xarray
        xmin: float, optional
            The min x coordinate, or None for automatic
        ymin: float, optional
            The min y coordinate, or None for automatic
        xmax: float, optional
            The max x coordinate, or None for automatic
        ymax: float, optional
            The max y coordinate, or None for automatic
        z: float, optional
            The z coordinate of the plane
        xspace: float, optional
            The extra space in x direction, before and after wind farm
        yspace: float, optional
            The extra space in y direction, before and after wind farm
        normalize_x: float, optional
            Divide x by this value
        normalize_y: float, optional
            Divide y by this value
        normalize_z: float, optional
            Divide z by this value
        normalize_v: dict, optional
            Divide the variables by these values
        label_map: dict
            The mapping from original to new field names
        vmin: dict
            Minimal values for variables
        vmax: dict
            Maximal values for variables
        states_sel: list, optional
            Reduce to selected states
        states_isel: list, optional
            Reduce to the selected states indices
        weight_turbine: int, optional
            Index of the turbine from which to take the weight
        to_file: str, optional
            Write data to this file name
        write_pars: dict
            Additional write function parameters
        ret_states: bool
            Flag for returning states indices
        ret_grid: bool
            Flag for returning grid data
        verbosity: int, optional
            The verbosity level, 0 = silent
        kwargs: dict, optional
            Parameters forwarded to the algorithm's calc_points
            function.

        Returns
        -------
        data: dict or pandas.DataFrame or xarray.Dataset
            The gridded data
        states: numpy.ndarray, optional
            The states indices
        grid_data: tuple, optional
            The grid data (x_pos, y_pos, z_pos, g_pts)

        """
        gdata = grids.get_grid_xy(
            self.fres, 
            resolution, 
            xmin, 
            ymin, 
            xmax, 
            ymax, 
            z, 
            xspace, 
            yspace, 
            verbosity-self.verbosity_delta,
        )

        data = self._calc_mean_data(
            "xy",
            data_format,
            variables,
            *gdata,
            normalize_x,
            normalize_y,
            normalize_z,
            normalize_v,
            label_map,
            vmin,
            vmax,
            states_sel,
            states_isel,
            weight_turbine,
            to_file,
            write_pars,
            ret_states,
            verbosity,
            **kwargs,
        )

        if ret_grid:
            out = list(data) if ret_states else [data]
            return tuple(out + [gdata])
        return data

    def get_mean_data_xz(
        self,
        resolution,
        variables=None,
        data_format="xarray",
        x_direction=270,
        xmin=None,
        zmin=0.0,
        xmax=None,
        zmax=None,
        y=None,
        xspace=500.0,
        zspace=500.0,
        normalize_x=None,
        normalize_y=None,
        normalize_z=None,
        normalize_v={},
        label_map={},
        vmin={},
        vmax={},
        states_sel=None,
        states_isel=None,
        weight_turbine=0,
        to_file=None,
        write_pars={},
        ret_states=False,
        ret_grid=False,
        verbosity=0,
        **kwargs,
    ):
        """
        Creates mean data of 2D farm flow slices in an xz-plane.

        Parameters
        ----------
        resolution: float
            The resolution in m
        variables: list of str, optional
            The variables, or None for all
        data_format: str
            The output data format: numpy, pandas, xarray
        x_direction: float, optional
            The direction of the x axis, 0 = north
        xmin: float, optional
            The min x coordinate, or None for automatic
        zmin: float, optional
            The min z coordinate
        xmax: float, optional
            The max x coordinate, or None for automatic
        zmax: float, optional
            The max z coordinate, or None for automatic
        y: float, optional
            The y coordinate of the plane
        xspace: float, optional
            The extra space in x direction, before and after wind farm
        zspace: float, optional
            The extra space in z direction, below and above wind farm
        normalize_x: float, optional
            Divide x by this value
        normalize_y: float, optional
            Divide y by this value
        normalize_z: float, optional
            Divide z by this value
        normalize_v: dict, optional
            Divide the variables by these values
        label_map: dict
            The mapping from original to new field names
        vmin: dict
            Minimal values for variables
        vmax: dict
            Maximal values for variables
        states_sel: list, optional
            Reduce to selected states
        states_isel: list, optional
            Reduce to the selected states indices
        weight_turbine: int, optional
            Index of the turbine from which to take the weight
        to_file: str, optional
            Write data to this file name
        write_pars: dict
            Additional write function parameters
        ret_states: bool
            Flag for returning states indices
        ret_grid: bool
            Flag for returning grid data
        verbosity: int, optional
            The verbosity level, 0 = silent
        kwargs: dict, optional
            Parameters forwarded to the algorithm's calc_points
            function.

        Returns
        -------
        data: dict or pandas.DataFrame or xarray.Dataset
            The gridded data
        states: numpy.ndarray, optional
            The states indices
        grid_data: tuple, optional
            The grid data (x_pos, y_pos, z_pos, g_pts)

        """
        gdata = grids.get_grid_xz(
            self.fres,
            resolution,
            x_direction,
            xmin,
            zmin,
            xmax,
            zmax,
            y,
            xspace,
            zspace,
            verbosity-self.verbosity_delta,
        )
        gdatb = (gdata[0], gdata[2], gdata[1], gdata[3])

        data = self._calc_mean_data(
            "xz",
            data_format,
            variables,
            *gdatb,
            normalize_x,
            normalize_z,
            normalize_y,
            normalize_v,
            label_map,
            vmin,
            vmax,
            states_sel,
            states_isel,
            weight_turbine,
            to_file,
            write_pars,
            ret_states,
            verbosity,
            **kwargs,
        )

        if ret_grid:
            out = list(data) if ret_states else [data]
            return tuple(out + [gdata])
        return data

    def get_mean_data_yz(
        self,
        resolution,
        variables=None,
        data_format="xarray",
        x_direction=270,
        ymin=None,
        zmin=0.0,
        ymax=None,
        zmax=None,
        x=None,
        yspace=500.0,
        zspace=500.0,
        normalize_x=None,
        normalize_y=None,
        normalize_z=None,
        normalize_v={},
        label_map={},
        vmin={},
        vmax={},
        states_sel=None,
        states_isel=None,
        weight_turbine=0,
        to_file=None,
        write_pars={},
        ret_states=False,
        ret_grid=False,
        verbosity=0,
        **kwargs,
    ):
        """
        Creates mean data of 2D farm flow slices in a yz-plane.

        Parameters
        ----------
        resolution: float
            The resolution in m
        variables: list of str, optional
            The variables, or None for all
        data_format: str
            The output data format: numpy, pandas, xarray
        x_direction: float, optional
            The direction of the x axis, 0 = north
        ymin: float, optional
            The min y coordinate, or None for automatic
        zmin: float, optional
            The min z coordinate
        ymax: float, optional
            The max y coordinate, or None for automatic
        zmax: float, optional
            The max z coordinate, or None for automatic
        x: float, optional
            The x coordinate of the plane
        yspace: float, optional
            The extra space in y direction, before and after wind farm
        zspace: float, optional
            The extra space in z direction, below and above wind farm
        normalize_x: float, optional
            Divide x by this value
        normalize_y: float, optional
            Divide y by this value
        normalize_z: float, optional
            Divide z by this value
        normalize_v: dict, optional
            Divide the variables by these values
        label_map: dict
            The mapping from original to new field names
        vmin: dict
            Minimal values for variables
        vmax: dict
            Maximal values for variables
        states_sel: list, optional
            Reduce to selected states
        states_isel: list, optional
            Reduce to the selected states indices
        weight_turbine: int, optional
            Index of the turbine from which to take the weight
        to_file: str, optional
            Write data to this file name
        write_pars: dict
            Additional write function parameters
        ret_states: bool
            Flag for returning states indices
        ret_grid: bool
            Flag for returning grid data
        verbosity: int, optional
            The verbosity level, 0 = silent
        kwargs: dict, optional
            Parameters forwarded to the algorithm's calc_points
            function.

        Returns
        -------
        data: dict or pandas.DataFrame or xarray.Dataset
            The gridded data
        states: numpy.ndarray, optional
            The states indices
        grid_data: tuple, optional
            The grid data (x_pos, y_pos, z_pos, g_pts)

        """
        gdata = grids.get_grid_yz(
            self.fres,
            resolution,
            x_direction,
            ymin,
            zmin,
            ymax,
            zmax,
            x,
            yspace,
            zspace,
            verbosity-self.verbosity_delta,
        )
        gdatb = (gdata[1], gdata[2], gdata[0], gdata[3])

        data = self._calc_mean_data(
            "yz",
            data_format,
            variables,
            *gdatb,
            normalize_y,
            normalize_z,
            normalize_x,
            normalize_v,
            label_map,
            vmin,
            vmax,
            states_sel,
            states_isel,
            weight_turbine,
            to_file,
            write_pars,
            ret_states,
            verbosity,
            **kwargs,
        )

        if ret_grid:
            out = list(data) if ret_states else [data]
            return tuple(out + [gdata])
        return data

    def _calc_states_data(
        self,
        ori,
        data_format,
        variables,
        a_pos,
        b_pos,
        c_pos,
        g_pts,
        normalize_a,
        normalize_b,
        normalize_c,
        normalize_v,
        label_map,
        vmin,
        vmax,
        states_sel,
        states_isel,
        to_file,
        write_pars,
        ret_states,
        verbosity,
        **kwargs,
    ):
        """Helper function for states data calculation"""
        # calculate point results:
        if states_sel is not None:
            kwargs["sel"] = {FC.STATE: states_sel}
        if states_isel is not None:
            kwargs["isel"] = {FC.STATE: states_isel}
        point_results = grids.calc_point_results(
            algo=self.algo,
            farm_results=self.fres,
            g_pts=g_pts,
            verbosity=verbosity-self.verbosity_delta,
            **kwargs,
        )
        states = point_results[FC.STATE].to_numpy()
        if variables is None:
            variables = list(point_results.data_vars.keys())
        else:
            point_results.drop_vars(variables)
        del g_pts

        # convert to numpy:
        data = {v: point_results[v].to_numpy() for v in variables}
        del point_results

        # apply data modification:
        a_pos, b_pos, c_pos, data = self._data_mod(
            a_pos,
            b_pos,
            c_pos,
            data,
            normalize_a,
            normalize_b,
            normalize_c,
            normalize_v,
            vmin,
            vmax,
        )

        # translate to selected format:
        if data_format == "numpy":
            data = grids.np2np_sp(data, states, a_pos, b_pos)
            self._write(data_format, data, to_file, verbosity, **write_pars)
        elif data_format == "pandas":
            data = grids.np2pd_sp(data, states, a_pos, b_pos, ori, label_map)
            self._write(data_format, data, to_file, verbosity, **write_pars)
        elif data_format == "xarray":
            data = grids.np2xr_sp(data, states, a_pos, b_pos, c_pos, ori, label_map)
            self._write(data_format, data, to_file, verbosity, **write_pars)
        else:
            raise ValueError(
                f"Unknown data format '{data_format}', choices: numpy, pandas, xarray"
            )

        return (data, states) if ret_states else data

    def get_states_data_xy(
        self,
        resolution,
        variables=None,
        data_format="xarray",
        xmin=None,
        ymin=None,
        xmax=None,
        ymax=None,
        z=None,
        xspace=500.0,
        yspace=500.0,
        normalize_x=None,
        normalize_y=None,
        normalize_z=None,
        normalize_v={},
        label_map={},
        vmin={},
        vmax={},
        states_sel=None,
        states_isel=None,
        to_file=None,
        write_pars={},
        ret_states=False,
        ret_grid=False,
        verbosity=0,
        **kwargs,
    ):
        """
        Creates states data of 2D farm flow slices in a horizontal xy-plane.

        Parameters
        ----------
        resolution: float
            The resolution in m
        variables: list of str, optional
            The variables, or None for all
        data_format: str
            The output data format: numpy, pandas, xarray
        xmin: float, optional
            The min x coordinate, or None for automatic
        ymin: float, optional
            The min y coordinate, or None for automatic
        xmax: float, optional
            The max x coordinate, or None for automatic
        ymax: float, optional
            The max y coordinate, or None for automatic
        z: float, optional
            The z coordinate of the plane
        xspace: float, optional
            The extra space in x direction, before and after wind farm
        yspace: float, optional
            The extra space in y direction, before and after wind farm
        normalize_x: float, optional
            Divide x by this value
        normalize_y: float, optional
            Divide y by this value
        normalize_z: float, optional
            Divide z by this value
        normalize_v: dict, optional
            Divide the variables by these values
        label_map: dict
            The mapping from original to new field names
        vmin: dict
            Minimal values for variables
        vmax: dict
            Maximal values for variables
        states_sel: list, optional
            Reduce to selected states
        states_isel: list, optional
            Reduce to the selected states indices
        to_file: str, optional
            Write data to this file name
        write_pars: dict
            Additional write function parameters
        ret_states: bool
            Flag for returning states indices
        ret_grid: bool
            Flag for returning grid data
        verbosity: int, optional
            The verbosity level, 0 = silent
        kwargs: dict, optional
            Parameters forwarded to the algorithm's calc_points
            function.

        Returns
        -------
        data: dict or pandas.DataFrame or xarray.Dataset
            The gridded data
        states: numpy.ndarray, optional
            The states indices
        grid_data: tuple, optional
            The grid data (x_pos, y_pos, z_pos, g_pts)

        """
        gdata = grids.get_grid_xy(
            self.fres, 
            resolution, 
            xmin, 
            ymin, 
            xmax, 
            ymax, 
            z, 
            xspace, 
            yspace, 
            verbosity-self.verbosity_delta,
        )

        data = self._calc_states_data(
            "xy",
            data_format,
            variables,
            *gdata,
            normalize_x,
            normalize_y,
            normalize_z,
            normalize_v,
            label_map,
            vmin,
            vmax,
            states_sel,
            states_isel,
            to_file,
            write_pars,
            ret_states,
            verbosity,
            **kwargs,
        )

        if ret_grid:
            out = list(data) if ret_states else [data]
            return tuple(out + [gdata])
        return data

    def get_states_data_xz(
        self,
        resolution,
        variables=None,
        data_format="xarray",
        x_direction=270,
        xmin=None,
        zmin=0.0,
        xmax=None,
        zmax=None,
        y=None,
        xspace=500.0,
        zspace=500.0,
        normalize_x=None,
        normalize_y=None,
        normalize_z=None,
        normalize_v={},
        label_map={},
        vmin={},
        vmax={},
        states_sel=None,
        states_isel=None,
        to_file=None,
        write_pars={},
        ret_states=False,
        ret_grid=False,
        verbosity=0,
        **kwargs,
    ):
        """
        Creates states data of 2D farm flow slices in an xz-plane.

        Parameters
        ----------
        resolution: float
            The resolution in m
        variables: list of str, optional
            The variables, or None for all
        data_format: str
            The output data format: numpy, pandas, xarray
        x_direction: float, optional
            The direction of the x axis, 0 = north
        xmin: float, optional
            The min x coordinate, or None for automatic
        zmin: float, optional
            The min z coordinate
        xmax: float, optional
            The max x coordinate, or None for automatic
        zmax: float, optional
            The max z coordinate, or None for automatic
        y: float, optional
            The y coordinate of the plane
        xspace: float, optional
            The extra space in x direction, before and after wind farm
        zspace: float, optional
            The extra space in z direction, below and above wind farm
        normalize_x: float, optional
            Divide x by this value
        normalize_y: float, optional
            Divide y by this value
        normalize_z: float, optional
            Divide z by this value
        normalize_v: dict, optional
            Divide the variables by these values
        label_map: dict
            The mapping from original to new field names
        vmin: dict
            Minimal values for variables
        vmax: dict
            Maximal values for variables
        states_sel: list, optional
            Reduce to selected states
        states_isel: list, optional
            Reduce to the selected states indices
        to_file: str, optional
            Write data to this file name
        write_pars: dict
            Additional write function parameters
        ret_states: bool
            Flag for returning states indices
        ret_grid: bool
            Flag for returning grid data
        verbosity: int, optional
            The verbosity level, 0 = silent
        kwargs: dict, optional
            Parameters forwarded to the algorithm's calc_points
            function.

        Returns
        -------
        data: dict or pandas.DataFrame or xarray.Dataset
            The gridded data
        states: numpy.ndarray, optional
            The states indices
        grid_data: tuple, optional
            The grid data (x_pos, y_pos, z_pos, g_pts)

        """
        gdata = grids.get_grid_xz(
            self.fres,
            resolution,
            x_direction,
            xmin,
            zmin,
            xmax,
            zmax,
            y,
            xspace,
            zspace,
            verbosity-self.verbosity_delta,
        )
        gdatb = (gdata[0], gdata[2], gdata[1], gdata[3])

        data = self._calc_states_data(
            "xz",
            data_format,
            variables,
            *gdatb,
            normalize_x,
            normalize_z,
            normalize_y,
            normalize_v,
            label_map,
            vmin,
            vmax,
            states_sel,
            states_isel,
            to_file,
            write_pars,
            ret_states,
            verbosity,
            **kwargs,
        )

        if ret_grid:
            out = list(data) if ret_states else [data]
            return tuple(out + [gdata])
        return data

    def get_states_data_yz(
        self,
        resolution,
        variables=None,
        data_format="xarray",
        x_direction=270,
        ymin=None,
        zmin=0.0,
        ymax=None,
        zmax=None,
        x=None,
        yspace=500.0,
        zspace=500.0,
        normalize_x=None,
        normalize_y=None,
        normalize_z=None,
        normalize_v={},
        label_map={},
        vmin={},
        vmax={},
        states_sel=None,
        states_isel=None,
        to_file=None,
        write_pars={},
        ret_states=False,
        ret_grid=False,
        verbosity=0,
        **kwargs,
    ):
        """
        Creates states data of 2D farm flow slices in a yz-plane.

        Parameters
        ----------
        resolution: float
            The resolution in m
        variables: list of str, optional
            The variables, or None for all
        data_format: str
            The output data format: numpy, pandas, xarray
        x_direction: float, optional
            The direction of the x axis, 0 = north
        ymin: float, optional
            The min y coordinate, or None for automatic
        zmin: float, optional
            The min z coordinate
        ymax: float, optional
            The max y coordinate, or None for automatic
        zmax: float, optional
            The max z coordinate, or None for automatic
        x: float, optional
            The x coordinate of the plane
        yspace: float, optional
            The extra space in y direction, before and after wind farm
        zspace: float, optional
            The extra space in z direction, below and above wind farm
        normalize_x: float, optional
            Divide x by this value
        normalize_y: float, optional
            Divide y by this value
        normalize_z: float, optional
            Divide z by this value
        normalize_v: dict, optional
            Divide the variables by these values
        label_map: dict
            The mapping from original to new field names
        vmin: dict
            Minimal values for variables
        vmax: dict
            Maximal values for variables
        states_sel: list, optional
            Reduce to selected states
        states_isel: list, optional
            Reduce to the selected states indices
        to_file: str, optional
            Write data to this file name
        write_pars: dict
            Additional write function parameters
        ret_states: bool
            Flag for returning states indices
        ret_grid: bool
            Flag for returning grid data
        verbosity: int, optional
            The verbosity level, 0 = silent
        kwargs: dict, optional
            Parameters forwarded to the algorithm's calc_points
            function.

        Returns
        -------
        data: dict or pandas.DataFrame or xarray.Dataset
            The gridded data
        states: numpy.ndarray, optional
            The states indices
        grid_data: tuple, optional
            The grid data (x_pos, y_pos, z_pos, g_pts)

        """
        gdata = grids.get_grid_yz(
            self.fres,
            resolution,
            x_direction,
            ymin,
            zmin,
            ymax,
            zmax,
            x,
            yspace,
            zspace,
            verbosity-self.verbosity_delta,
        )
        gdatb = (gdata[1], gdata[2], gdata[0], gdata[3])

        data = self._calc_states_data(
            "yz",
            data_format,
            variables,
            *gdatb,
            normalize_y,
            normalize_z,
            normalize_x,
            normalize_v,
            label_map,
            vmin,
            vmax,
            states_sel,
            states_isel,
            to_file,
            write_pars,
            ret_states,
            verbosity,
            **kwargs,
        )

        if ret_grid:
            out = list(data) if ret_states else [data]
            return tuple(out + [gdata])
        return data
